with Ada.Text_IO, Reps, Linear, Fouriers_Init, Qs.IO, Qs.Ops;
use  Ada.Text_IO, Reps, Linear, Fouriers_Init;

generic

  Number,Size,PartSize: in Positive;
  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Scal(R1,R2: Rep) return Scalar is <>;
  with function Ball(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function Approx(S: Scalar) return Rep is <>;
  with function Inf(S: Scalar) return Rep is <>;
  with function Sup(S: Scalar) return Rep is <>;
  with function InfAbs(S: Scalar) return Rep is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Envelope(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with function "*"(R: Rep; S: Scalar) return Scalar is <>;
  with function "/"(S: Scalar; R: Rep) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Widen(S: Scalar) return Scalar is <>;
  with procedure Show(S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Get(F: in File_Type; S: out Scalar; Decimal: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := True) is <>;

package Basics is

   package Qb is new Qs (Number => Number, Size => Size, Scalar => Scalar);
   package Qb_IO is new Qb.IO;
   package Qb_Ops is new Qb.Ops;
   use Qb;

   FullMax: constant Positive := Number*Size;
   PartMax: constant Positive := Number*PartSize;

   subtype Freq is Qb.Components.Freq;
   subtype FullIndex is Integer range 1 .. FullMax;  -- vector index
   subtype PartIndex is Integer range 1 .. PartMax;  -- vector index

   function F(Alpha: Scalar; H: Q) return Q;                -- F(H) := (alpha/4)(-D^(-2)H +D^(-3) H^2)
   function DF(Alpha: Scalar; H,G: Q) return Q;             -- DF(H)G
   function Utilde(Alpha,Lambda: Scalar; H,G: Q) return Q;  -- P2(F(H))+Lambda*G
   function DUtilde(Alpha: Scalar; H0,H,G: Q) return Q;     -- P2(DF(H0)H)
   function W(Alpha: Scalar; H0,H,G: Q) return Q;           -- (T-P2*T*P1)H , where T=DF(H0)-I+c*P1
   function W1(Alpha: Scalar; H: Q) return Q;               -- -H-(alpha/4)D^(-2)H
   function W2(Alpha: Scalar; H0,H,G: Q) return Q;          -- W(H)-W1(H)
   function W2P1(Alpha: Scalar; H0,H,G: Q) return Q;        -- W2(P1(H))
   function W2P2(Alpha: Scalar; H0,H,G: Q) return Q;        -- W2(P2(H))

--- for numerical experiments

   function L(Alpha: Scalar; H,G: Q) return Q;              -- approx -(4*D^4+alpha*D^2)G+2*alpha*D(G*H)
   function DF(Alpha: Scalar; H: Q) return Matrix;          -- approx DF(H)
   function DFmI(Alpha: Scalar; H: Q) return Matrix;        -- approx (DF(H)-I)^(-1)+I
   function L(Alpha: Scalar; H: Q) return Matrix;           -- approx L
   function DUtilde(Alpha: Scalar; H,G: Q) return Matrix;   -- approx P2*DF(H)
   function DUtildemI(Alpha: Scalar; H,G: Q) return Matrix; -- approx I+1/(M-I) , where M=DUtilde(H)
   function Wi(Alpha: Scalar; H0,G: Q) return Matrix;       -- approx 1/(T-P2*T*P1) , where T=DFF(H0)+c*P1

end Basics;
