with Ada.Text_IO, Ada.Integer_Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Ada.Integer_Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
with Intervals, Intervals.Ops, Intervals.IO;
use  Intervals, Intervals.Ops, Intervals.IO;
with Ks;

package body Checkbif is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Interval;

   DecIO:  constant Boolean  := False;     -- decimal input/output for Scalar
   Number: constant Positive := 1;

   procedure Show(St: in String; Sc: in Scalar) is
   begin
      Show(St,False); Show(Sc);
   end Show;

   procedure Check_PF(Size, PartSize: in Positive; QName: in String;
                   DA,DL,R1,R2,R3,R4,R5: in Radius; Div: in Integer) is

      package CBKs is new Ks (Number=>Number,Size => Size,PartSize => PartSize,Scalar => Scalar);
      package CBKs_Qf renames CBKs.Qf;
      package CBKs_Qf_Ops renames CBKs.Qf_Ops;
      package CBKs_Qf_IO renames CBKs.Qf_IO;
      package CBKs_Qf_SLinear renames CBKs.FSLinear;
      use CBKs, CBKs_Qf, CBKs_Qf_Ops, CBKs_Qf_IO, CBKs_Qf_SLinear;

      S: Integer := 1;
      A,Alpha,Alpha0,G03,G11,G01wc,G02wc,G01w,G00en,G01ec,G00es: Scalar;
      H0,H,G: Q;
      Wi0 : Matrix(PartIndex,PartIndex);
      DDL: Radius := DL/Radius(Div);

      function Gder(Nd,Np: Integer; Alpha,Lambda: Scalar; R: Radius; NewH: Boolean := True) return Scalar is
         Err: constant Radius := 1.0E-17;
      begin
         if NewH then
            H := H0;
            NewtonU(Rep(S)*Alpha,Lambda,G,Err,H);
            if not UT(Rep(S)*Alpha,Lambda,H,G,R) then
               Show("UT in Gder fails");
            end if;
            if Sup(Norm(H-H0)) > R5 then
               Show("GDer balls not in big ball!!");
            end if;
         end if;
         return Rep(S**Nd)*Gder(Nd,Np,Rep(S)*Alpha,A,Wi0,H,G);
      end Gder;

   begin
      Read("data/" & QName,1,Alpha0,H0,False);
      G := ApproxBifVec(Alpha0,H0);
      Alpha := Widen(Alpha0+Ball(DA));
      if not UT(Alpha,Ball(DL),H0,G,R5,3,10) then
         Show("UT on big ball fails");
      end if;
      if UT(Alpha0,Scal(0),H0,G,R1) then
         H0 := Ball(H0,R1);
      else
         Show("UT on small ball fails");
         Show("lambda",Approx(P1(H0,G)));
         NewtonU(Alpha0,Scal(Zero),G,1.0E-17,H0);
         Write(QName & "-updated",Alpha0,H0,False);
      end if;
      Wi0 := Wi(Alpha0,H0,G);
      A := WiWNorm(Alpha,Wi0,H0,G);

--    full rectange, in Div pieces
      G03 := Gder(0,3,Alpha,Piece(1,Div,Ball(DL)),R2);
      G11 := Gder(1,1,Alpha,Piece(1,Div,Ball(DL)),R2,False);
      for I in 2 .. Div loop
         G03 := Envelope(G03,Gder(0,3,Alpha,Piece(I,Div,Ball(DL)),R2));
         G11 := Envelope(G11,Gder(1,1,Alpha,Piece(I,Div,Ball(DL)),R2,False));
      end loop;
      if Inf(G03*G11) > Zero then
         Show("replacing g(alpha,lambda) by g(-alpha,lambda)");
         Alpha0 := -Alpha0;
         S := -1;
         G11 := -G11;
      end if;
--      Show("g03(.,.) = ",G03);
--      Show("g11(.,.) = ",G11);

--    west side
      G01wc := Gder(0,1,Alpha0-Scal(DA),Scal(0),R3);
      G02wc := Gder(0,2,Alpha0-Scal(DA),Scal(0),R3,False);
      G01w  := G01wc+Half*DL*Ball(G02wc);
--      Show("g01(-,.) = ",G01w);

--    east side
      G00en := Gder(0,0,Alpha0+Scal(DA), Scal(DL),R4);  -- north
      G00es := Gder(0,0,Alpha0+Scal(DA),-Scal(DL),R4);  -- south
      G01ec := Gder(0,1,Alpha0+Scal(DA), Scal(0),R3);   -- center
--      Show("g00(+,+) = ",G00en);
--      Show("g00(+,-) = ",G00es);
--      Show("g01(+,0) = ",G01ec);

      if InfAbs(G03)    > Zero then Show("condition 1  OK"); else S := 0; end if;
      if Sup(G03*G11)   < Zero then Show("condition 2  OK"); else S := 0; end if;
      if Inf(G03*G01w)  > Zero then Show("condition 3  OK"); else S := 0; end if;
      if Inf(G03*G00en) > Zero then Show("condition 4+ OK"); else S := 0; end if;
      if Sup(G03*G00es) < Zero then Show("condition 4- OK"); else S := 0; end if;
      if Sup(G03*G01ec) < Zero then Show("condition 5  OK"); else S := 0; end if;

      if S =  1 then Show("Pitchfork bifurcation."); end if;
      if S = -1 then Show("Inverse pitchfork bifurcation."); end if;
      if S =  0 then Show("Something is wrong"); end if;
   end Check_PF;

   procedure Check_Int(Size, PartSize: in Positive; QName: in String; DA,DL,R1,R2,R3: in Radius) is
      --   R1: radius in UT, center
      --   R2: radius in UT, full rectangle
      --   R3: radius in UT, other points

      package CBKs is new Ks (Number=>Number,Size => Size,PartSize => PartSize,Scalar => Scalar);
      package CBKs_Qf renames CBKs.Qf;
      package CBKs_Qf_Ops renames CBKs.Qf_Ops;
      package CBKs_Qf_IO renames CBKs.Qf_IO;
      package CBKs_Qf_SLinear renames CBKs.FSLinear;
      use CBKs, CBKs_Qf, CBKs_Qf_Ops, CBKs_Qf_IO, CBKs_Qf_SLinear;

      pragma Suppress(Storage_Check);

      A,Alpha0,Lambda0: Scalar;
      G02,G11,G01wc,G01ec: Scalar;
      H0,H,G: Q;
      Wi0 : Matrix(PartIndex,PartIndex);
      OK: Boolean := True;

      function Gder(Nd,Np: Integer; Alpha,Lambda: Scalar; R: Radius; NewH: Boolean := True) return Scalar is
         Err: constant Radius := 1.0E-17;
      begin
         if NewH then
            H := H0;
            NewtonU(Alpha,Lambda,G,Err,H);
            if not UT(Alpha,Lambda,H,G,R) then
               Show("UT in Gder fails");
            end if;
         end if;
         return Gder(Nd,Np,Alpha,A,Wi0,H,G);
      end Gder;

   begin
      Read("data/" & QName,1,Alpha0,H0,False);
      G := ApproxBifVec(Alpha0,H0);
      Lambda0 := P1(H0,G);
      if UT(Alpha0,Lambda0,H0,G,R1) then
         H0 := Ball(H0,R1);
      else
         Show("UT on small ball fails");
      end if;
      Wi0 := Wi(Alpha0,H0,G);
      A := WiWNorm(Alpha0+Ball(DA),Wi0,H0,G);

--    full rectange
      G02 := Gder(0,2,Alpha0+Ball(DA),Lambda0+Ball(DL),R2);
      G11 := Gder(1,1,Alpha0+Ball(DA),Lambda0+Ball(DL),R2,False);
--      Show("g02(.,.) = ",G02);
--      Show("g11(.,.) = ",G11);

--    west side
      G01wc := Gder(0,1,Alpha0-Scal(DA),Lambda0,R3);
--      Show("g01(-,0) : ",G01wc);

--    east side
      G01ec := Gder(0,1,Alpha0+Scal(DA),Lambda0,R3);
--      Show("g01(+,0) = ",G01ec);

      if InfAbs(G02) > Zero then Show("condition 1  OK"); else OK := False; end if;
      if InfAbs(G11) > Zero then Show("condition 2  OK"); else OK := False; end if;
      if Sup(G01wc*G01ec) < Zero then Show("condition 3  OK"); else OK := False; end if;

--      Show("g00(+,+) = ",Gder(0,0,Alpha0+Scal(DA),Lambda0+Scal(DL),R3));
--      Show("g00(+,-) = ",Gder(0,0,Alpha0+Scal(DA),Lambda0-Scal(DL),R3));
--      Show("g00(-,+) = ",Gder(0,0,Alpha0-Scal(DA),Lambda0+Scal(DL),R3));
--      Show("g00(-,-) = ",Gder(0,0,Alpha0-Scal(DA),Lambda0-Scal(DL),R3));

      if Inf(G02*Gder(0,0,Alpha0+Scal(DA),Lambda0+Scal(DL),R3)) > Zero then
         Show("condition 4  OK");
      else
         OK := False;
      end if;
      if Inf(G02*Gder(0,0,Alpha0+Scal(DA),Lambda0-Scal(DL),R3)) > Zero then
         Show("condition 5  OK");
      else
         OK := False;
      end if;
      if Inf(G02*Gder(0,0,Alpha0-Scal(DA),Lambda0+Scal(DL),R3)) > Zero then
         Show("condition 6  OK");
      else
         OK := False;
      end if;
      if Inf(G02*Gder(0,0,Alpha0-Scal(DA),Lambda0-Scal(DL),R3)) > Zero then
         Show("condition 7  OK");
      else
         OK := False;
      end if;

      if OK then
         Show("Intersection bifurcation");
      else
         Show("Something is wrong");
      end if;
   end Check_Int;

   procedure Check_Fold(Size, PartSize: in Positive; QName: in String;
                        DA,DL,R1,R2,R3,R4: in Radius; Div: in Integer) is

      package CBKs is new Ks (Number=>Number,Size => Size,PartSize => PartSize,Scalar => Scalar);
      package CBKs_Qf renames CBKs.Qf;
      package CBKs_Qf_Ops renames CBKs.Qf_Ops;
      package CBKs_Qf_IO renames CBKs.Qf_IO;
      package CBKs_Qf_SLinear renames CBKs.FSLinear;
      use CBKs, CBKs_Qf, CBKs_Qf_Ops, CBKs_Qf_IO, CBKs_Qf_SLinear;

      pragma Suppress(Storage_Check);

      A,Alpha0,Lambda0: Scalar;
      G02,G10,G00wc,G01wc,G00w,G00en,G00ec,G00es: Scalar;
      H0,H,G: Q;
      Wi0 : Matrix(PartIndex,PartIndex);
      S: Integer := 1;

      function Gder(Nd,Np: Integer; Alpha,Lambda: Scalar; R: Radius; NewH: Boolean := True) return Scalar is
         Err: constant Radius := 1.0E-17;
      begin
         if NewH then
            H := H0;
            NewtonU(Alpha,Lambda,G,Err,H);
            if not UT(Alpha,Lambda,H,G,R) then
               Show("UT in Gder fails");
            end if;
         end if;
         return Gder(Nd,Np,Alpha,A,Wi0,H,G);
      end Gder;

   begin
      Read("data/" & QName,1,Alpha0,H0,DecIO);
      G := ApproxBifVec(Alpha0,H0);
      Lambda0 := P1(H0,G);
      if not UT(Alpha0+Ball(DA),Lambda0+Ball(DL),H0,G,R4) then
         Show("UT on big ball fails");
      end if;
      if UT(Alpha0,Lambda0,H0,G,R1) then
         H0 := Ball(H0,R1);
      else
         Show("UT on small ball fails");
      end if;
      Wi0 := Wi(Alpha0,H0,G);
      A := WiWNorm(Alpha0+Ball(DA),Wi0,H0,G);

--    full rectange
      G02 := Gder(0,2,Alpha0+Ball(DA),Lambda0+Piece(1,Div,Ball(DL)),R2);
      for I in 2 .. Div loop
         G02 := Envelope(G02,Gder(0,2,Alpha0+Ball(DA),Lambda0+Piece(I,Div,Ball(DL)),R2));
      end loop;
--      Show("g02(.,.) = ",G02);
      G10 := Gder(1,0,Alpha0+Ball(DA),Lambda0+Ball(DL),R2,False);
--      Show("g10(.,.) = ",G10);

--    west side
      G00wc := Gder(0,0,Alpha0-Scal(DA),Lambda0,R3);
      G01wc := Gder(0,1,Alpha0-Scal(DA),Lambda0,R3,False);
      G00w  := G00wc+DL*Ball(G01wc);
--      Show("g00(-,.) : ",G00w);

--    east side
      G00en := Gder(0,0,Alpha0+Scal(DA),Lambda0+Scal(DL),R3);  -- north
      G00es := Gder(0,0,Alpha0+Scal(DA),Lambda0-Scal(DL),R3);  -- south
      G00ec := Gder(0,0,Alpha0+Scal(DA),Lambda0,R3);           -- center
--      Show("g00(+,+) = ",G00en);
--      Show("g00(+,-) = ",G00es);
--      Show("g00(+,0) = ",G00ec);

      if InfAbs(G02)    > Zero then Show("condition 1  OK"); else S := 0; end if;
      if Sup(G02*G10)   < Zero then Show("condition 2  OK"); else S := 0; end if;
      if Inf(G02*G00w)  > Zero then Show("condition 3  OK"); else S := 0; end if;
      if Inf(G02*G00en) > Zero then Show("condition 4+ OK"); else S := 0; end if;
      if Inf(G02*G00es) > Zero then Show("condition 4- OK"); else S := 0; end if;
      if Sup(G02*G00ec) < Zero then Show("condition 5  OK"); else S := 0; end if;
      if S=1 then
         Show("Fold bifurcation");
      else
         Show("Something is wrong");
      end if;
   end Check_Fold;

   procedure Check_FoldR(Size, PartSize: in Positive; QName: in String;
                         DA,DL,R1,R2,R3,R4: in Radius; Div: in Integer) is

      package CBKs is new Ks (Number=>Number,Size => Size,PartSize => PartSize,Scalar => Scalar);
      package CBKs_Qf renames CBKs.Qf;
      package CBKs_Qf_Ops renames CBKs.Qf_Ops;
      package CBKs_Qf_IO renames CBKs.Qf_IO;
      package CBKs_Qf_SLinear renames CBKs.FSLinear;
      use CBKs, CBKs_Qf, CBKs_Qf_Ops, CBKs_Qf_IO, CBKs_Qf_SLinear;

      pragma Suppress(Storage_Check);

      S: Integer := 1;
      A,Alpha0,Lambda0: Scalar;
      G02,G10,G00wc,G01wc,G00w,G00en,G00ec,G00es: Scalar;
      H0,H,G: Q;
      Wi0 : Matrix(PartIndex,PartIndex);

      function Gder(Nd,Np: Integer; Alpha,Lambda: Scalar;
                    R: Radius; NewH: Boolean := True) return Scalar is
         Err: constant Radius := 1.0E-17;
      begin
         if NewH then
            H := H0;
            NewtonU(Rep(S)*Alpha,Lambda,G,Err,H);
            if not UT(Rep(S)*Alpha,Lambda,H,G,R) then
               Show("UT in Gder fails");
            end if;
         end if;
        return Rep(S**Nd)*Gder(Nd,Np,Rep(S)*Alpha,A,Wi0,H,G);
      end Gder;

   begin
      Read("data/" & QName,1,Alpha0,H0,DecIO);
      G := ApproxBifVec(Alpha0,H0);
      Lambda0 := P1(H0,G);
      if not UT(Alpha0+Ball(DA),Lambda0+Ball(DL),H0,G,R4) then
         Show("UT on big ball fails");
      end if;
      if UT(Alpha0,Lambda0,H0,G,R1) then
         H0 := Ball(H0,R1);
      else
         Show("UT on small ball fails");
      end if;

      Wi0 := Wi(Alpha0,H0,G);
      A := WiWNorm(Alpha0+Ball(DA),Wi0,H0,G);
--      Show("replacing g(alpha,lambda) by g(-alpha,lambda)");
      Alpha0 := -Alpha0;
      S := -1;

--    full rectange
      G02 := Gder(0,2,Alpha0+Ball(DA),Lambda0+Piece(1,Div,Ball(DL)),R2);
      G10 := Gder(1,0,Alpha0+Ball(DA),Lambda0+Piece(1,Div,Ball(DL)),R2,False);
      for I in 2 .. Div loop
         G02 := Envelope(G02,Gder(0,2,Alpha0+Ball(DA),Lambda0+Piece(I,Div,Ball(DL)),R2));
         G10 := Envelope(G10,Gder(1,0,Alpha0+Ball(DA),Lambda0+Piece(I,Div,Ball(DL)),R2,False));
      end loop;
--      Show("g02(.,.) = ",G02);
--      Show("g10(.,.) = ",G10);

--    west side
      G00wc := Gder(0,0,Alpha0-Scal(DA),Lambda0,R3);
      G01wc := Gder(0,1,Alpha0-Scal(DA),Lambda0,R3,False);
      G00w  := G00wc+DL*Ball(G01wc);
--      Show("g00(-,.) : ",G00w);

--    east side
      G00en := Gder(0,0,Alpha0+Scal(DA),Lambda0+Scal(DL),R3);  -- north
      G00es := Gder(0,0,Alpha0+Scal(DA),Lambda0-Scal(DL),R3);  -- south
      G00ec := Gder(0,0,Alpha0+Scal(DA),Lambda0,R3);           -- center
--      Show("g00(+,+) = ",G00en);
--      Show("g00(+,-) = ",G00es);
--      Show("g00(+,0) = ",G00ec);

      if InfAbs(G02)    > Zero then Show("condition 1  OK"); else S := 0; end if;
      if Sup(G02*G10)   < Zero then Show("condition 2  OK"); else S := 0;  end if;
      if Inf(G02*G00w)  > Zero then Show("condition 3  OK"); else S := 0;  end if;
      if Inf(G02*G00en) > Zero then Show("condition 4+ OK"); else S := 0;  end if;
      if Inf(G02*G00es) > Zero then Show("condition 4- OK"); else S := 0;  end if;
      if Sup(G02*G00ec) < Zero then Show("condition 5  OK"); else S := 0;  end if;
      if S=-1 then
         Show("Inverse fold bifurcation");
      else
         Show("Something is wrong");
      end if;
   end Check_FoldR;

end Checkbif;
