with Ada.Text_IO, Messages, Reps.Ops, Reps.IO, Linear, Linear.IO;
use Ada.Text_IO, Messages, Reps.Ops, Reps.IO, Linear, Linear.IO;
with Intervals, Intervals.Ops, Intervals.IO;
use  Intervals, Intervals.Ops, Intervals.IO;
with Ks;

procedure CheckBranch(Size, PartSize: in Positive; QName1: in String;
                      R1,DA1,DL1: in Radius; QName2: in String;
                      R2,DA2,DL2: in Radius; DLStart: in Rep; Err: in Radius) is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);

   subtype Scalar is Interval;
   Number: Positive := 1;

   package IKS is new Ks (Number=>Number,Size => Size,PartSize => PartSize,Scalar => Scalar);
   package IKS_Qf renames IKS.Qf;
   package IKS_Qf_Ops renames IKS.Qf_Ops;
   package IKS_Qf_IO renames IKS.Qf_IO;
   package IKS_Qf_SLinear renames IKS.FSLinear;
   use IKS, IKS_Qf, IKS_Qf_Ops, IKS_Qf_IO, IKS_Qf_SLinear;

   type Counter is mod 200;

   DecIO:    constant Boolean  := False;            -- decimal i/o for Scalar
   RA0:      constant Radius   := One/Rep(16);      -- large radius of analyticity
   RA1:      constant Radius   := Rep(15)/Rep(256); -- small radius of analyticity
   RSmax:    constant Radius   := 1.0E-10;          -- max radius of the small ball
   DFmin:    constant Radius   := 1.0E-24;          -- smallest frequency increment
   OnePlus:  constant Radius   := 1.01;
   OneMinus: constant Radius   := 0.8;
   LessOne:  constant Radius   := 0.999;

   Count:    Counter := 0;
   DeltaCIM: Integer := 10;
   RL:       Radius  := 1.0E-4;          -- initial radius of the large ball
   DF:       Radius  := 1.0E-12;         -- initial frequency increment
   FreqLast: Rep;                        -- used if endpoint is not a bifurcation

   UTok,DAok,DLok,BBok: Boolean;
   CheckInvMan: Integer;
   FreqFirst,Freq0,Freq1: Rep;
   RS0,RS1 : Radius;
   A,Alpha1,Alpha2,Alpha3,AlphaFirst,Lambda: Scalar;
   G,H0,H1,H2,H,H3: Q;
   EV: Matrix(PartIndex,PartIndex);
   EVi: SMatrix(PartIndex,PartIndex);
   TT,SSi: BlockOp;
   F: File_Type;

begin
   if QName2 = "" then
      Create(F,Out_File,"output/" & Qname1 & "-80");
      FreqLast := R2;
   elsif QName2 = "64iP3p" or QName2 = "64iP3n" then
      Create(F,Out_File,"output/" & Qname2);
      FreqLast := R2;
   else
      Read("data/" & QName2,1,Alpha3,H3,False);
      FreqLast  := Inf(Alpha3)-LessOne*DA2;
      Create(F,Out_File,"output/" & Qname1 & "-" & Qname2);
   end if;

   Read("data/" & QName1,1,Alpha1,H1,False);
   SetRho(RA0,H1);
   CheckInvMan := Integer(Inf(Alpha1)/Rep(DeltaCIM)+Half)*DeltaCIM;

   if DA1>Zero then
      G := ApproxBifVec(Alpha1,H1);
      Alpha2 := Scal(Approx(Alpha1)+LessOne*DA1);
      Lambda := P1(H1,G);
      H2 := H1;
      NewtonU(Alpha2,Lambda+Scal(DLStart),G,Err,H2);
      Newton(Alpha2,Err,H2);   -- computes starting point of the branch
      H0 := H2;
      AlphaFirst := Alpha2;
      FreqFirst := Inf(AlphaFirst);
      Txt_Put(F,FreqFirst);
      Txt_Put(F,Approx(Norm(H2)));
      New_Line(F);
      RS0 := RSMax;
      USmall(FreqFirst,H0,RS0);
      UTok := UT(Alpha1+Ball(DA1),Lambda+Ball(DL1),H1,G,R1,3,10);
      DLok := SupAbs(Lambda-P1(Ball(H2,RS0),G)) < DL1;
      BBok := Sup(Norm(H1-Ball(H2,RS0))) < R1;
      if not UTok then Show("UT on big ball fails"); end if;
      if not DLok then Show("DL too large"); end if;
      if not BBok then Show("Not in big ball"); end if;
   else
      H0 := H1;
      FreqFirst := Approx(Alpha1);
      RS0 := RSMax;
      USmall(FreqFirst,H0,RS0);
   end if;

   Freq0 := FreqFirst;
   SetInterval(Freq0,RS0,Err,DF,RL,H0);
   Show("First DF:",DF);
   RS1 := RS0;

   while (Freq0 < FreqLast) and (DF>DFmin) loop
      Freq1 := Freq0+DF;
      H := H0;
      Newton(Scal(Freq0,Freq1),Err,H);
      if Sup(Norm(H-H0)+Scal(RS0))<RL then
         H1 := H;
         if Freq1 > Rep(CheckInvMan) then
            Freq1 :=  Rep(CheckInvMan);
            Newton(Scal(Freq1),Err,H1);
            USmall(Freq1,H1,RS1);
            Show("Check at",CheckInvMan);
            CheckInvMan := CheckInvMan+DeltaCIM;
            MakeLinear(Scal(Freq1),H1,SSi,TT,EV,EVi);
            A := ASiNorm(Scal(Freq1),EVi,Ball(H1,RS1),D(Ball(H1,RS1),RA1),EV,TT,SSi);
            if Sup(A)<One then Show(" OK."); else Show(" failed."); end if;
            Show("norm= ",False);
            Show(Norm(H1)+Ball(RS1));
            Show("L2 norm= ",False);
            Show(L2Norm(H1));
         else
            Newton(Scal(Freq1),Err,H1);
            USmall(Freq1,H1,RS1);
         end if;
         if Sup(Norm(H-H1)+Scal(RS1))<RL then
            if U(Scal(Freq0,Freq1),H,RL) then
               H0 := H1;
               RS0 := RS1;
               Freq0 := Freq1;
               DF := OnePlus*DF;
            else
               SetInterval(Freq0,RS0,Err,DF,RL,H0);
            end if;
         else
            DF := OneMinus*DF;
            BigR(Freq0,DF,RS0,RL,H0);
         end if;
      else
         DF := OneMinus*DF;
         BigR(Freq0,DF,RS0,RL,H0);
      end if;
      Count := Count+1;
      if Count=0 then
         Txt_Put(F,Freq1);
--         Txt_Put(F,Approx(Norm(H1)));
--         Txt_Put(F,Approx(L2Norm(H1)));
         Txt_Put(F,FreqLast-Freq1);
         New_Line(F);
--         Write(Qname1 & "-" & Qname2,Scal(Freq1),H1,False);
      end if;
   end loop;

   if Freq0 < FreqLast then Show("incomplete"); end if;

   if QName2 = "" then
      Show("Last Freqency:",Freq1,15);
      Write("output/" & Qname1 & "-final",Scal(Freq1),H1,False);
   elsif QName2 = "64iP3p" or QName2 = "64iP3n" then
      Write("data/" & Qname2,Scal(Freq1),H1,False);
   else
      H2 := H0;
      G := ApproxBifVec(Alpha3,H3);
      Lambda := P1(H3,G);
      UTok := UT(Alpha3+Ball(DA2),Lambda+Ball(DL2),H3,G,R2,3,10);
      DLok := SupAbs(Lambda-P1(Ball(H2,RS0),G))  < DL2;
      DAok := SupAbs(Alpha3-Scal(Freq1)) < DA2;
      BBok := Sup(Norm(H3-Ball(H2,RS0))) < R2;
      if not UTok then Show("UT on big ball fails"); end if;
      if not DLok then Show("DL too large"); end if;
      if not DAok then Show("DA too large"); end if;
      if not BBok then Show("Not in big ball"); end if;
      if UTok and DAok and DLok and BBok then
         Show("Ending point ok");
      end if;
   end if;

   Close(F);

end CheckBranch;
