with Ada.Text_IO, Messages,Reps, Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Messages,Reps, Linear, Linear.IO, Linear.Ops;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Find_Bif is

-- this program can be used to approximately locate a bifurcation
-- starting from the given approximate solution, it moves along
-- a branch of solutions by TotD, then tries to locate a bifurcation
-- by looking for a zero eigenvalue

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;
   Size:     constant Positive := GetArg(1); -- Maximal frequency for explicit Fourier polynomials
   QName:    constant String   := GetArg(2); -- filename of approximate solution
   TotD:     constant Rep      := GetArg(3); -- frequency delta
   Steps:    constant Integer  := GetArg(4); -- steps in frequency
   DecIO:    constant Boolean  := GetArg(5); -- decimal input/output for Scalar

   PartSize: constant Positive := Size;      -- Cutoff for matrix approximations
   R1:       constant Radius   := 0.1;       -- radius of analyticity
   Err:      constant Radius   := 1.0E-14;   -- error tolerated on Newton
   Err2:     constant Radius   := 1.0E-5;    -- error tolerated on zero EV

   Number: Positive;
   D: Scalar := Scal(TotD)/Scal(Steps);
   F: File_Type;
   Alpha: Scalar;

begin
   Open(F,In_File,QName);
   Get(F,Alpha,DecIO);
   Txt_Int_IO.Get(F,Number);
   Close(F);

   declare
      package FBKs is new Ks (Number => Number, Size => Size,
                              PartSize => PartSize, Scalar => Scalar);
      use FBKs, FBKs.Qf, FBKs.Qf_Ops, FBKs.Qf_IO;

      H: Q;

   begin
      Read(QName,1,Alpha,H);
      for I in 0 .. Steps loop
         Newton(Alpha,Err,H);
         Alpha := Alpha+D;
      end loop;
      Focus(0.1*Approx(D),Err,Err2,Alpha,H);
      Show(Alpha);
   end;

end Find_Bif;
