with Ada.Text_IO, Messages, Reps, Linear, Linear.IO;
use Ada.Text_IO, Messages, Reps, Linear, Linear.IO;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Find_Br is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;

   Size:     constant Positive := GetArg(1);    -- Maximal frequency for explicit Fourier polynomials
   PartSize: constant Positive := GetArg(2);    -- Cutoff for matrix approximations
   QName:    constant String   := GetArg(3);    -- filename of approximate solution
   FreqLast: constant Rep      := GetArg(4);    -- last frequency
   Steps:    constant Integer  := GetArg(5);    -- steps in frequency
   DA:       constant Rep      := GetArg(6);    -- delta alpha
   DL:       constant Rep      := GetArg(7);    -- delta lambda

   Number:   constant Positive := 1;

   package IKS is new Ks (Number=>Number, Size => Size, PartSize => PartSize, Scalar => Scalar);
   package IKS_Qf renames IKS.Qf;
   package IKS_Qf_Ops renames IKS.Qf_Ops;
   package IKS_Qf_IO renames IKS.Qf_IO;
   package IKS_Qf_SLinear renames IKS.FSLinear;
   use IKS, IKS_Qf, IKS_Qf_Ops, IKS_Qf_IO, IKS_Qf_SLinear;

   pragma Suppress(Storage_Check);

   Err: constant Radius := 1.0E-17;

   Alpha,AlphaInterval,AlphaFirst,AlphaLast: Scalar;
   G,H: Q;

begin

   Read(QName,1,Alpha,H,False);
   G := ApproxBifVec(Alpha,H);
   AlphaFirst := Alpha+Scal(DA);
   NewtonU(AlphaFirst,P1(H,G)+Scal(DL),G,Err,H);
   Newton(AlphaFirst,Err,H);
   AlphaLast     := Scal(FreqLast);
   AlphaInterval := (AlphaLast-AlphaFirst)/Scal(Steps);

   for I in 1 .. Steps loop
      Alpha := AlphaFirst+Scal(I)*AlphaInterval;
      Newton(Alpha,Err,H);
   end loop;
   Newton(AlphaLast,Err,H);
   Write(QName & "-final",AlphaLast,H,False);

end Find_Br;
