with Ada.Text_IO, Ada.Integer_Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Ada.Integer_Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Find_Int is

-- this program can be used to approximately locate an intersection bifurcation
-- starting from the given approximate solution
-- it provides the value of the function g(alpha,lambda) in the corners of the
-- rectancle IxJ, as described in Lemma 5.2

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;

   Size:     constant Positive := GetArg(1); -- Maximal frequency for explicit Fourier polynomials
   PartSize: constant Positive := GetArg(2); -- Cutoff for matrix approximations
   QName:    constant String   := GetArg(3); -- filename of approximate solution
   DA:       constant Radius   := GetArg(4); -- half-width of the interval I in Lemma 5.2
   DL:       constant Radius   := GetArg(5); -- half-width of the interval J in Lemma 5.2
   DecIO:    constant Boolean  := False;     -- uses hex in/out

   Number: Positive;
   Alpha0: Scalar;
   F: File_Type;

begin

   Open(F,In_File,QName);
   Get(F,Alpha0,DecIO);
   Txt_Int_IO.Get(F,Number);
   Close(F);

   declare
      package CBKs is new Ks (Number=>Number,Size => Size, PartSize => PartSize,Scalar => Scalar);
      package CBKs_Qf renames CBKs.Qf;
      package CBKs_Qf_Ops renames CBKs.Qf_Ops;
      package CBKs_Qf_IO renames CBKs.Qf_IO;
      package CBKs_Qf_SLinear renames CBKs.FSLinear;
      use CBKs, CBKs_Qf, CBKs_Qf_Ops, CBKs_Qf_IO, CBKs_Qf_SLinear;

      pragma Suppress(Storage_Check);

      Err: constant Radius := 1.0E-17;

      A,Lambda0: Scalar;
      H0,G: Q;
      Wi0: Matrix(PartIndex,PartIndex);

      function Gder(Nd,Np: Integer; Alpha,Lambda: Scalar) return Scalar is
      begin
         NewtonU(Alpha,Lambda,G,Err,H0);
         return Gder(Nd,Np,Alpha,A,Wi0,H0,G);
      end Gder;

      procedure Show(St: in String; Sc: in Scalar) is
      begin
        Show(St,False); Show(Sc);
      end Show;

   begin
      Read(QName,1,Alpha0,H0,DecIO);
      G := ApproxBifVec(Alpha0,H0);
      Lambda0 := P1(H0,G);
      Wi0 := Wi(Alpha0,H0,G);
      A := WiWNorm(Alpha0+Ball(DA),Wi0,H0,G);

      Show("g00(+,+) = ",Gder(0,0,Alpha0+Scal(DA),Lambda0+Scal(DL)));
      Show("g00(+,-) = ",Gder(0,0,Alpha0+Scal(DA),Lambda0-Scal(DL)));
      Show("g00(-,+) = ",Gder(0,0,Alpha0-Scal(DA),Lambda0+Scal(DL)));
      Show("g00(-,-) = ",Gder(0,0,Alpha0-Scal(DA),Lambda0-Scal(DL)));
   end;

end Find_Int;

