with Ada.Text_IO, Messages, Reps,Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Focus_PF is

-- This program can be used to approximately locate a pitchfork bifurcation.
-- Starting from the given approximate solution, it moves along
-- a branch of solutions until alpha=A1, then tries to locate a bifurcation
-- in the interval [A1,A2]

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;

   Size:     constant Positive := GetArg(1); -- Maximal frequency for explicit Fourier polynomials
   PartSize: constant Positive := GetArg(2); -- Cutoff for matrix approximations
   QName:    constant String   := GetArg(3); -- filename of approximate solution
   A1:       Rep               := GetArg(4); -- lower freqency
   A2:       Rep               := GetArg(5); -- higher freqency
   Steps:    constant Integer  := GetArg(6); -- steps used to reach alpha=A1
   DecIO:    constant Boolean  := False;     -- uses hex input/output
   Max_Iter: constant Integer  := 50;        -- iterations in the procedure
   R1:       constant Radius   := 0.1;       -- radius of analyticity
   Err:      constant Radius   := 1.0E-15;   -- error tolerated on Newton

   Number: Positive;
   F: File_Type;
   Alpha,A,Temp,T1,T2: Scalar;

begin
   Open(F,In_File,QName);
   Get(F,Alpha,DecIO);
   Txt_Int_IO.Get(F,Number);
   Close(F);

   declare
      package FBKs is new Ks (Number => Number, Size => Size, PartSize => PartSize, Scalar => Scalar);
      use FBKs, FBKs.Qf, FBKs.Qf_Ops, FBKs.Qf_IO;

      H,G: Q;
      Wi0 : Matrix(PartIndex,PartIndex);
      D: Rep := (A1-Approx(Alpha))/Rep(Steps);

      function Gprime(Alpha: Scalar) return Scalar is
      begin
         NewtonU(Alpha,Scal(Zero),G,Err,H);
         return Gder(0,1,Alpha,A,Wi0,H,G);
      end Gprime;

   begin
      Read(QName,1,Alpha,H,DecIO);
      for I in 0 .. Steps loop
         Newton(Alpha,Err,H);
         Alpha := Alpha+Scal(D);
      end loop;

      G := ApproxBifVec(Alpha,H);
      Wi0 := Wi(Alpha,H,G);
      A := WiWNorm(Alpha,Wi0,H,G);
      Show("G'(nu,0):    ",False);
      T1 := Gprime(Alpha);
      Show(T1);

      Alpha := Scal(A2);
      Newton(Alpha,Err,H);
      G := ApproxBifVec(Alpha,H);
      Wi0 := Wi(Alpha,H,G);
      A := WiWNorm(Alpha,Wi0,H,G);
      Show("G'(nu,0):    ",False);
      T2 := Gprime(Alpha);
      Show(T2);

      if T1*T2 < Scal(Zero) then
         for I in 1 .. Max_Iter loop
            Alpha := Scal(0.5*(A1+A2));
            Newton(Alpha,Err,H);
            G := ApproxBifVec(Alpha,H);
            Wi0 := Wi(Alpha,H,G);
            A := WiWNorm(Alpha,Wi0,H,G);
            Temp := Gprime(Alpha);
            if T1*Temp>Scal(0) then
               A1 := Approx(Alpha);
            else
               A2 := Approx(Alpha);
            end if;
         end loop;
         Show("G'(nu,0):    ",False);
         Show(Temp);
         Show(Alpha);
         Write(QName & "-final",Alpha,H,DecIO);
      else
         Show("Unsuccessful.");
      end if;
   end;

end Focus_PF;
