
package body Fouriers.IO is

  package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
  use Txt_Int_IO;

  procedure Show(H: in Fourier; Cut: in Rep := Zero) is
  begin
     Put("Parity" & H.J'Img);
     New_Line;
     Put("Rho");
     Show(Scal(H.R));
     Put_Line("Coefficients");
     for L in Freq loop
        if Cut<SupAbs(Component(L,H)) then
           Put(Standard_Output,L,3,10);
           if L<Par(H) then
              Put(" !!! should be zero >>> ");
           end if;
           Put("  ");
           Show(Component(L,H));
        end if;
     end loop;
     if Poly(H) then return; end if;
     Put_Line("Errors ");
     for L in FreqE loop
        if Cut<SupAbs(ErrComponent(L,H)) then
           Put(Standard_Output,L,3,10);
           if L<Par(H) then
              Put(" !!! should be zero >>> ");
           end if;
           Put("  ");
           Show(ErrComponent(L,H));
        end if;
     end loop;
  end Show;

  procedure Put(F: in File_Type; H: in  Fourier; Decimal: in Boolean := True) is
  begin
     if Trunc then
        Put(F,-Size);
     else
        Put(F,Size);
     end if;
     New_Line(F);
     Txt_Put(F,Rho(H),Decimal);
     New_Line(F);
     for L in Par(H) .. Size loop
        Put(F,Component(L,H),Decimal);
        New_Line(F);
     end loop;
     if Trunc then return; end if;
     for L in Par(H) .. SizeE loop
        Put(F,ErrComponent(L,H),Decimal);
        New_Line(F);
     end loop;
  end Put;

  procedure Write(Name: in String; H: in Fourier; Decimal: in Boolean := True) is
    F: File_Type;
  begin
     Put_Line("Writing " & Name);
     Create(F,Out_File,Name);
     Put(F,H,Decimal);
     Close(F);
  end Write;

  procedure Get(F: in File_Type; J: in Parity; H: out Fourier; Decimal: in Boolean := True) is
    Temp : Scalar;
    N: Integer;
    R: Radius;
    OK : Boolean;
  begin
    Get(F,N);
    OK := Trunc or (N<0) or not Decimal;
    Txt_Get(F,R,Decimal);
    SetZero(J,R,H);
    for L in J .. Abs(N) loop
       Get(F,Temp,Decimal);
       if Temp=ScalZero then
          null;
       elsif L<=Size then
          H.C(L) := Temp;
       else
          OK := Trunc;
       end if;
    end loop;
    if N>0 then
       for L in J .. 2*N loop
          Get(F,Temp,Decimal);
          if Temp=ScalZero then
             null;
          elsif L<=SizeE then
             H.E(L) := Temp;
          else
             OK := Trunc;
          end if;
       end loop;
    end if;
    if not OK then
       Put_Line("Warning: Input truncated or inaccurate");
    end if;
    H.P := False;
    H.P := Poly(H,True);
  end Get;

  procedure Read(Name: in String; J: in Parity; H: out Fourier; Decimal: in Boolean := True) is
    F: File_Type;
  begin
    Put_Line("Reading " & Name);
    Open(F,In_File,Name);
    Get(F,J,H,Decimal);
    Close(F);
  end Read;

end Fouriers.IO;
