with Messages;
use Messages;


package body Fouriers is

   function Par(H: Fourier) return Parity is
   begin
      return H.J;
   end Par;

   function Rho(H: Fourier) return Radius is
   begin
      return H.R;
   end Rho;

   procedure SetRho(R: in Radius; H: in out Fourier) is
   begin
      H.R := R;
   end SetRho;

   function Poly(H: Fourier; Check: in Boolean := False) return Boolean is
   begin
     if Trunc then return True; end if;
     if not Check then return H.P; end if;
     for I in FreqE loop
       if H.E(I) /= ScalZero then
         if H.P then raise Constraint_Error; end if;
         return False;
       end if;
     end loop;
     return True;
   end Poly;

   procedure SetZero(J: in Parity; R: in Radius; H: out Fourier) is
   begin
      H.P := True;
      H.J := J;
      H.R := R;
      H.C := ZeroFouCoeffs;
      if not Trunc then H.E := ZeroFouErrors; end if;
   end SetZero;

   function IsZero(H: Fourier) return Boolean is
   begin
      for N in Freq loop
         if H.C(N) /= ScalZero then return False; end if;
      end loop;
      return Poly(H,True);
   end IsZero;

   function Component(N: Freq; H: Fourier) return Scalar is
   begin
      return H.C(N);
   end Component;

   function ErrComponent(N: FreqE; H: Fourier) return Scalar is
   begin
      if Poly(H) then return ScalZero; end if;
      return H.E(N);
   end ErrComponent;

   procedure SetComponent(N: Freq; S: in Scalar; H: in out Fourier) is
   begin
     if N<H.J and then S /= ScalZero then
       Message("Fouriers.SetComponent",Index_Error);
     end if;
     H.C(N) := S;
   end SetComponent;

   procedure SetErrComponent(N: FreqE; S: in Scalar; H: in out Fourier) is
   begin
      if not (Trunc or else S=ScalZero) then
         H.P := False;
         H.E(Integer'Max(N,H.J)) := Abs(S);
      end if;
   end SetErrComponent;

begin

  for N in Freq loop ZeroFouCoeffs(N) := ScalZero; end loop;
  for N in FreqE loop ZeroFouErrors(N) := ScalZero; end loop;

end Fouriers;
