with Reps, Fouriers_Init;
use Reps, Fouriers_Init;

generic

  Size: in Positive;

  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Ball(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function Approx(S: Scalar) return Rep is <>;
  with function Inf(S: Scalar) return Rep is <>;
  with function Sup(S: Scalar) return Rep is <>;
  with function InfAbs(S: Scalar) return Rep is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Envelope(S1,S2: Scalar) return Scalar is <>;

package Fouriers is

  pragma Elaborate_Body;

  SizeE:  constant Positive := 2*Size;

  subtype Freq  is Integer range 0 .. Size;
  subtype FreqE is Integer range 0 .. SizeE;
  type Fourier is private;

  function Par(H: Fourier) return Parity;                         -- Parity of H  (0=even 1=odd)
  function Rho(H: Fourier) return Radius;                         -- Weight defining domain of H
  procedure SetRho(R: in Radius; H: in out Fourier);              -- Set Weight defining domain of H
  function Poly(H: Fourier; Check: in Boolean := False) return Boolean; -- H is a Fourier polynomial
  procedure SetZero(J: in Parity; R: in Radius; H: out Fourier);        -- H := 0
  function IsZero(H: Fourier) return Boolean;                           -- True if H=0
  function Component(N: Freq; H: Fourier) return Scalar;                -- H[N]
  function ErrComponent(N: FreqE; H: Fourier) return Scalar;            -- H_err[N]
  procedure SetComponent(N: Freq; S: in Scalar; H: in out Fourier);     -- H[N] := S
  procedure SetErrComponent(N: FreqE; S: in Scalar; H: in out Fourier); -- H_err[N] := S

private

  type ScalVec is array(Freq) of Scalar;
  type ScalVecE is array(FreqE) of Scalar;

  type Fourier is
    record
      P: Boolean;      --- if True, we have a Fourier polynomial
      J: Parity;       --- even=0, odd=1
      R: Radius;       --- domain parameter
      C: ScalVec;      --- coeffs
      E: ScalVecE;     --- these coeffs are also functions of q
    end record;

  ScalZero: constant Scalar  := Scal(Zero);
  ScalOne:  constant Scalar  := Scal(One);
  ScalHalf: constant Scalar  := Scal(Half);
  Trunc:    constant Boolean := IsNumeric(ScalZero);

  ZeroFouCoeffs: ScalVec;
  ZeroFouErrors: ScalVecE;

end Fouriers;
