with Messages, Reps.Ops, Linear.IO, Linear.Ops, Packed, Rounding;
use Messages, Reps.Ops, Linear.IO, Linear.Ops;

package body Ks is

   use Qf_Ops, Qf.Components, Qf_Ops.Components_Ops;

   PartMax: constant Positive := Number*PartSize;
   Four:         constant Rep := Two*Two;
   ScalZero:  constant Scalar := Scal(Zero);
   ScalOne:   constant Scalar := Scal(One);
   Trunc:    constant Boolean := IsNumeric(ScalZero);

   procedure SetSubComp(M: in Freq; N: in Part; S: in Scalar; H: in out Q) renames Qf.SetSubComponent;
   procedure SetErrComp(M: FreqE; N: Part; S: in Scalar; H: in out Q) renames Qf.SetErrComponent;

   pragma Inline (SetSubComp,SetErrComp);

   procedure ShowParam is
   begin
     if Trunc then
        Show("Running Numeric version");
     else
        Show("Running Interval version");
     end if;
     Show("  Number =",Number);
     Show("    Size =",Size);
     Show("PartSize =",PartSize);
   end ShowParam;

   function FF(Alpha: Scalar; H: Q) return Q is
   begin
     return F(Alpha,H)-H;
   end FF;

   function DFF(Alpha: Scalar; H,G: Q) return Q is
   begin
      return DF(Alpha,H,G)-G;
   end DFF;

   function Fdot(H: Q) return Q is
     C: constant Scalar := Scal(Half*Half);
   begin
      return Dinv(3,C,ZeroConstant(H*H))-Dinv(2,C,H);
   end Fdot;

   function DFdot(H,G: Q) return Q is
   begin
      return Dinv(3,Scal(Half),ZeroConstant(H*G))-Dinv(2,Scal(Half*Half),G);
   end DFdot;

   function D2F(Alpha: Scalar; G1,G2: Q) return Q is
   begin
      return Dinv(3,Half*Alpha,ZeroConstant(G1*G2));
   end D2F;

   function Contr(Alpha: Scalar; M: Matrix; H: Q) return Q is
      G: Q;
   begin
      G := F(Alpha,H);
      return G-M*(G-H);
   end Contr;

   function DContr(Alpha: Scalar; M: Matrix; H,G: Q) return Q is
      F: Q;
   begin
      F := DF(Alpha,H,G);
      return F-M*(F-G);
   end DContr;

   function DCNorm(Alpha: Scalar; M: Matrix; H: Q) return Scalar is
      MidSize: constant Integer := Integer'Min(PartSize+10,Size);
      NTest: Integer;
      S,E: Rep;
      S1,S2: Scalar;
      G1,G2: Q;
   begin
      TraceEnter("Ks.DCNorm");
      S1 := ScalZero;
      for M1 in Part loop
         S2 := ScalZero;
         for M2 in Part loop
            S := Zero;
            NTest := MidSize;
            for N2 in Par(H) .. Size loop
               SetZero(Par(H),Rho(H),G2);
               SetSubComp(N2,M2,ScalOne,G2);
               G1 := DContr(Alpha,M,H,G2);
              E := Sup(Norm(Component(M1,G1))/Norm(Component(M2,G2)));
--               Show("from deg =" & N2'Img & " get",E);
               S := Rep'Max(S,E);
               if N2=NTest then                              -- test higher order ball
                  exit when Trunc;
                  SetZero(Par(H),Rho(H),G2);
                  SetErrComp(N2+1,M2,ScalOne,G2);
                  G1 := DContr(Alpha,M,H,G2);
                 E := Sup(Norm(Component(M1,G1)));
                  if E<S or N2=Size then                     -- done
--                     Show("from deg >" & N2'Img & " get",E);
                     S := Rep'Max(S,E);
                     exit;
                  end if;
                  NTest := Integer'Min(NTest+10,Size);       -- for next test
               end if;
            end loop;
--            Show("M1 =" & M1'Img & " M2 =" & M2'Img & " add",S);
            S2 := S2+Scal(S);
         end loop;
--         Show("M1 =" & M1'Img & " totals",Sup(S2));
         S1 := Max(S1,S2);
      end loop;
--      Show("Max over M1 is",Sup(S1));
      TraceLeave;
      return S1;
   end DCNorm;

   function UContr(Alpha,Lambda: Scalar; M: Matrix; H,G: Q) return Q is
      G1: Q;
   begin
      G1 := Utilde(Alpha,Lambda,H,G);
      return G1-M*(G1-H);
   end UContr;

   function DUContr(Alpha: Scalar; M: Matrix; H0,H,G: Q) return Q is
      F: Q;
   begin
      F := DUtilde(Alpha,H0,H,G);
      return F-M*(F-H);
   end DUContr;

   function UNorm(Alpha: Scalar; M: Matrix; H,G: Q) return Scalar is
      MidSize: constant Integer := Integer'Min(PartSize+10,Size);
      NTest: Integer;
      S,E: Rep;
      S1,S2: Scalar;
      G1,G2: Q;
   begin
      TraceEnter("Ks.UNorm");
      S1 := ScalZero;
      for M1 in Part loop
         S2 := ScalZero;
         for M2 in Part loop
            S := Zero;
            NTest := MidSize;
            for N2 in Par(H) .. Size loop
               SetZero(Par(H),Rho(H),G2);
               SetSubComp(N2,M2,ScalOne,G2);
               G1 := DUContr(Alpha,M,H,G2,G);
               E := Sup(Norm(Component(M1,G1))/Norm(Component(M2,G2)));
--               Show("from deg =" & N2'Img & " get",E);
               S := Rep'Max(S,E);
               if N2=NTest then
                  exit when Trunc;
                  SetZero(Par(H),Rho(H),G2);
                  SetErrComp(N2+1,M2,ScalOne,G2);
                  G1 := DUContr(Alpha,M,H,G2,G);
                  E := Sup(Norm(Component(M1,G1)));
                  if E<S or N2=Size then                     -- done
                     S := Rep'Max(S,E);
                     exit;
                  end if;
                  NTest := Integer'Min(NTest+10,Size);       -- for next test
               end if;
            end loop;
            S2 := S2+Scal(S);
         end loop;
         S1 := Max(S1,S2);
      end loop;
      TraceLeave;
      return S1;
   end UNorm;

   function BlockSize(K: Positive; A: BlockOp; Check: Boolean := False) return Integer is
   begin
     if K>PartMax or else A.Im(K)=ScalZero then return 1; end if;
     if K<PartMax and then (A.Re(K+1)=A.Re(K) and A.Im(K+1)=-A.Im(K)) then return 2; end if;
     if Check then
       Message("Ks.BlockSize",Parameter_Error);
     end if;
     return 0;
   end BlockSize;

   procedure Convert(Ar,Ai: in Vector; A: out BlockOp) is
     K: Integer := 1;
   begin
     for L in PartIndex loop
       A.Re(L) := Scal(Ar(L));
       A.Im(L) := Scal(-Ai(L));
     end loop;
     while K <= PartMax loop
       K := K+BlockSize(K,A,True);   -- check
     end loop;
   end Convert;

   function Norm(A,B: BlockOp; R: Radius) return Scalar is
     K: Integer := 1;
     DK: Integer;
     Fac: constant Scalar := Exp(Scal(R));
     E: Scalar := ScalZero;
     S1,S2: Scalar;
   begin
     while K <= PartMax loop
       DK :=  BlockSize(K,A,True);
       S1 := Abs(A.Re(K));
       S2 := Abs(B.Re(K));
       if DK=2 then
          Show(A.Re(K));
          Show(A.Im(K));
         S1 := S1+Fac*Abs(A.Im(K));
       end if;
       E := Max(E,S1*Inv(S2));
       K := K+DK;
     end loop;
     return E;
   end Norm;

   function Inv(A: BlockOp) return BlockOp is
     K: Integer := 1;
     DK: Integer;
     S: Scalar;
     B: BlockOp;
   begin
     while K <= PartMax loop
       DK := BlockSize(K,A,True);
       if DK=1 then
         B.Re(K) := Inv(A.Re(K));
         B.Im(K) := ScalZero;
       else
         S := Sqr(A.Re(K))+Sqr(A.Im(K));
         B.Re(K) := A.Re(K)/S;
         B.Im(K) := -A.Im(K)/S;
         B.Re(K+1) := B.Re(K);
         B.Im(K+1) := -B.Im(K);
       end if;
       K := K+DK;
     end loop;
     return B;
   end Inv;

   function "*"(A,B: BlockOp) return BlockOp is
     K: Integer := 1;
     DK: Integer;
     C: BlockOp;
   begin
     while K <= PartMax loop
       DK := BlockSize(K,A,True);
       if DK=1 then
         C.Re(K) := A.Re(K)*B.Re(K);
         C.Im(K) := ScalZero;
       else
         C.Re(K) := A.Re(K)*B.Re(K)-A.Im(K)*B.Im(K);
         C.Im(K) := A.Re(K)*B.Im(K)+A.Im(K)*B.Re(K);
         C.Re(K+1) := C.Re(K);
         C.Im(K+1) := -C.Im(K);
       end if;
       if BlockSize(K,B,True) /= DK then
         Message("Ks.Prod",Parameter_Error);
       end if;
       K := K+DK;
     end loop;
     return C;
   end "*";

   function "*"(A: BlockOp; G: Q) return Q is
     K: Integer := 1;
     DK: Integer;
     N0,N1: Freq;
     M0,M1: Part;
     Cr,Ci: Scalar;
     H: Q;
   begin
     if not Poly(G) then
       Message("Ks.Prod",Not_Implemented);
     end if;
     SetZero(Par(G),Rho(G),H);
     while K <= PartMax loop
       DK := BlockSize(K,A,True);
       UnPack(K,N0,M0);
       if DK=1 then
         Cr := SubComponent(N0,M0,G);
         SetSubComponent(N0,M0,A.Re(K)*Cr,H);
       else
         UnPack(K+1,N1,M1);
         Cr := SubComponent(N0,M0,G);
         Ci := SubComponent(N1,M1,G);
         SetSubComponent(N0,M0,A.Re(K)*Cr-A.Im(K)*Ci,H);
         SetSubComponent(N1,M1,A.Im(K)*Cr+A.Re(K)*Ci,H);
       end if;
       K := K+DK;
     end loop;
     return H;
   end "*";

   function T(Alpha: Scalar; G: Q) return Q is
   begin
     return Scal(-4)*D(4,G)-Alpha*D(2,G);
   end T;

   function Tinv(Alpha: Scalar; G: Q) return Q is
   begin
     return Dinv(2,Alpha/Four,Scal(-One/Four),G);
   end Tinv;

   function DTinv(Alpha: Scalar; G: Q) return Q is
   begin
     return Dinv(1,Alpha/Four,Scal(-One/Four),G);
   end DTinv;

   function LmT(Alpha: Scalar; H,DH,G,DG: Q) return Q is
   begin
     return (Two*Alpha)*(G*DH+DG*H);
   end LmT;

   function L(Alpha: Scalar; H,DH,G: Q) return Q is
   begin
      if not Poly(G) then
         Message("Ks.L",Not_Implemented);
      end if;
     return LmT(Alpha,H,DH,G,D(1,G))+T(Alpha,G);
   end L;

   function InvTminusImagNorm(TT: BlockOp; K: Integer; R: Radius) return Scalar is
     Rho: constant Scalar := Exp(Scal(R));
     Larger1: Boolean;
     Tau,Sig,Sig2,C0,C1,C3,T: Scalar;
   begin
     Tau := TT.Re(K);
     Sig := abs(TT.Im(K)/Tau);
     Sig2 := Sqr(Sig);
     C0 := -Sig2*(Sig2+Scal(4));
     C1 := (-Two)*Rho*Sig*Sig2;
     C3 := Two*Rho*Sig;
     Root4(C0,C1,C3,One,T,Larger1);
     return (T+Rho*Sig)/(Tau*Sqrt(Sqr(Sqr(T)-Sig2)+Scal(4)*Sig2));
   end InvTminusImagNorm;

   procedure MakeLinear(Alpha: in Scalar; H: in Q; SSi,TT: out BlockOp; EV: out Matrix; EVinv: out SMatrix) is
     K: Integer := 1;
     S: Scalar;
     EvR,EvI: Vector(PartIndex);
     M,MR,MI: Matrix(PartIndex,PartIndex);
   begin
     TraceEnter("Ks.MakeLinear");
     M := L(Alpha,H);

     Rounding.Relax_Checks;
     Eigen(M,EvR,EvI,MR,MI);
     Rounding.Restore_Checks;

     EV := MR;
     Convert(EvR,EvI,TT);

     while K <= PartMax loop
       if BlockSize(K,TT)=1 then
         SSi.Re(K) := Inv(TT.Re(K));
         SSi.Im(K) := ScalZero;
         K := K+1;
       else
         S := InvTminusImagNorm(TT,K,Rho(H));
         if Sup(TT.Re(K))<Zero then S := -S; end if;   -- either sign is fine
         SSi.Re(K) := S;
         SSi.Im(K) := ScalZero;
         SSi.Re(K+1) := S;
         SSi.Im(K+1) := ScalZero;
         SetColumn(K+1,GetColumn(K,MI),EV);
         K := K+2;
       end if;
     end loop;

     EVinv := GetNumeric(EV);
     Invert(EVinv);
     Show("Number of EV with positive real part:",CountLarge(EvR));
     TraceLeave;
   end MakeLinear;

   function ASi(Alpha: Scalar; EVi: SMatrix; H,DH: Q; EV: Matrix; TT,SSi: BlockOp; G: Q) return Q is
     Gl,Gh,Fl,Fh: Q;
   begin
     Split(PartSize,G,Gl,Gh);
     Split(PartSize,LmT(Alpha,H,DH,Tinv(Alpha,Gh),DTinv(Alpha,Gh))+L(Alpha,H,DH,EV*(SSi*Gl)),Fl,Fh);
     return EVi*Fl+Fh-TT*(SSi*Gl);
   end ASi;

   function ASiNorm(Alpha: Scalar; EVi: SMatrix; H,DH: Q; EV: Matrix; TT,SSi: BlockOp) return Scalar is
      MidSize: constant Integer := Integer'Min(PartSize+10,Size);
      NTest: Integer;
      S,E: Rep;
      S1,S2: Scalar;
      G1,G2: Q;
   begin
      TraceEnter("Ks.ASiNorm");
      S1 := ScalZero;
      for M1 in Part loop
         S2 := ScalZero;
         for M2 in Part loop
            S := Zero;
            NTest := MidSize;
            for N2 in Par(H) .. Size loop
               SetZero(Par(H),Rho(H),G2);
               SetSubComp(N2,M2,ScalOne,G2);
               G1 := ASi(Alpha,EVi,H,DH,EV,TT,SSi,G2);
               E := Sup(Norm(Component(M1,G1))/Norm(Component(M2,G2)));
               S := Rep'Max(S,E);
               if N2=NTest then                           -- test higher order ball
                  exit when Trunc;
                  SetZero(Par(H),Rho(H),G2);
                  SetErrComp(N2+1,M2,ScalOne,G2);
                  G1 := ASi(Alpha,EVi,H,DH,EV,TT,SSi,G2);
                  E := Sup(Norm(Component(M1,G1)));
                  if E<S or N2=Size then                  -- done
                     S := Rep'Max(S,E);
                     exit;
                  end if;
                  NTest := Integer'Min(NTest+10,Size);    -- for next test
               end if;
            end loop;
            S2 := S2+Scal(S);
         end loop;
         S1 := Max(S1,S2);
      end loop;
      TraceLeave;
      return S1;
   end ASiNorm;

   function Wi1(Alpha: Scalar; H: Q) return Q is
   begin
     return Dinv(Alpha/Four,-ScalOne,H);
   end Wi1;

   function Wi3(Alpha: Scalar; Wi0: Matrix; H: Q) return Q is
     Hl,Hh: Q;
   begin
     Split(PartSize,H,Hl,Hh);
     return Wi0*Hl+Wi1(Alpha,Hh);
   end Wi3;

   function WiW(Alpha: Scalar; Wi0: Matrix; H0,H,G: Q) return Q is
     H1,H2,H3,H4: Q;
   begin
     H1 := W1(Alpha,H);
     H2 := Basic.W2P1(Alpha,H0,H,G);
     H3 := Basic.W2P2(Alpha,H0,H,G);
     Split(PartSize,H1+H2+H3,H1,H2);
     Split(PartSize,H-Wi0*H1,H1,H2);
     Split(PartSize,H3,H2,H4);
     return H1+Wi1(Alpha,H4);
   end WiW;

   function WiW(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q is
     E: Scalar;
     Ht: Q;
   begin
     Ht := H;
     SplitLowErr(PartSize,Ht,E);
     return Ball(WiW(Alpha,Wi0,H0,Ht,G),Sup(A*E));
   end WiW;

   function WiWNorm(Alpha: Scalar; Wi0: Matrix; H0,G: Q) return Scalar is
---  produce Wi0 with Wi
      MidSize: constant Integer := Integer'Min(PartSize+10,Size);
      NTest: Integer;
      S,E: Rep;
      S1,S2: Scalar;
      H1,H2: Q;
   begin
      TraceEnter("Ks.WiWNorm");
      S1 := ScalZero;
      for M1 in Part loop
         S2 := ScalZero;
         for M2 in Part loop
            S := Zero;
            NTest := MidSize;
            for N2 in Par(H0) .. Size loop
               SetZero(Par(H0),Rho(H0),H2);
               SetSubComp(N2,M2,ScalOne,H2);
               H1 := WiW(Alpha,Wi0,H0,H2,G);
               E := Sup(Norm(Component(M1,H1))/Norm(Component(M2,H2)));
               S := Rep'Max(S,E);
               if N2=NTest then                          -- test higher order ball
                  exit when Trunc;
                  SetZero(Par(H0),Rho(H0),H2);
                  SetErrComp(N2+1,M2,ScalOne,H2);
                  H1 := WiW(Alpha,Wi0,H0,H2,G);
                  E := Sup(Norm(Component(M1,H1)));
                  if E<S or N2=Size then                 -- done
                     S := Rep'Max(S,E);
                     exit;
                  end if;
                  NTest := Integer'Min(NTest+10,Size);   -- for next test
               end if;
            end loop;
            S2 := S2+Scal(S);
         end loop;
         S1 := Max(S1,S2);
      end loop;
      TraceLeave;
      return S1;
   end WiWNorm;

   function Wi(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q is
---  produce Wi0 with Wi, and A with WiWNorm
     M: constant Integer := 3;   -- controls precision
     E: Scalar;
     Hn,Hr: Q;
   begin
     if Sup(A) >= One then
       Show("A=",Sup(A));
       Message("Ks.Wi",Domain_Error);
     end if;
     Hn := Wi3(Alpha,Wi0,H);
     Hr := Hn;
     for N in 1 .. M-1 loop
       Hn := WiW(Alpha,A,Wi0,H0,Hn,G);
       Hr := Hr+Hn;
     end loop;
     if not Trunc then
       Hn := WiW(Alpha,A,Wi0,H0,Hn,G);
       E := Inv(ScalOne-A)*Norm(Hn);
       Hr := Ball(Hr,Sup(E));
     end if;
     return Hr;
   end Wi;

   function PWiP(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q is
   begin
     return P2(Wi(Alpha,A,Wi0,H0,P2(H,G),G),G);
   end PWiP;

   function UT(Alpha,Lambda: Scalar; H,G: Q; R: Radius; NA,NL: Positive := 1) return Boolean is
      AlphaI,LambdaJ,Eps,K : Scalar;
      M     : Matrix(PartIndex,PartIndex);
   begin
      for I in 1 .. NA loop
        AlphaI := Piece(I,NA,Alpha);
        M := DUtildemI(AlphaI,H,G);
        K := UNorm(AlphaI,M,Ball(H,R),G);
        for J in 1 .. NL loop
          LambdaJ := Piece(J,NL,Lambda);
          Eps := Norm(H-UContr(AlphaI,LambdaJ,M,H,G));
          if Inf(R*(Scal(One)-K)-Eps) <= Zero then
            Show("Eps =",Sup(Eps));
            Show("K   =",Sup(K));
            return False;
          end if;
        end loop;
      end loop;
      return True;
   end UT;

   function U(Alpha: Scalar; H: Q; R: Radius) return Boolean is
     M: constant Matrix(PartIndex,PartIndex) := DFmI(Alpha,H);
     Eps: constant Scalar := Norm(H-Contr(Alpha,M,H));
     K: constant Scalar :=  DCNorm(Alpha,M,Ball(H,R));
   begin
     return Inf(R*(Scal(One)-K)-Eps) > Zero;
   end U;

   function GG(Alpha: Scalar; H,G: Q) return Scalar is
   begin
      return P1(F(Alpha,H)-H,G);
   end GG;

   function Gder(Nd,Np: Integer; Alpha,A: Scalar; Wi0: Matrix; H0,G: Q) return Scalar is

     function Wi(F: Q) return Q is
     begin
       return Wi(Alpha,A,Wi0,H0,F,G);
     end Wi;

   begin
     if Nd=0 and Np=0 then return P1(F(Alpha,H0)-H0,G); end if;

     declare
       FD,FP,FDP,FPP,FPPP,HD,HPG: Q;
     begin
       FP   := Wi(DFF(Alpha,H0,G));
       if Nd=0 and Np=1 then return P1(FP,G); end if;
       FD   := Wi(Fdot(H0));
       if Nd=1 and Np=0 then return P1(FD,G); end if;
       HPG  := G-P2(FP,G);                             -- uhat+fprime
       FPP  := Wi(D2F(Alpha,HPG,HPG));
       if Nd=0 and Np=2 then return P1(FPP,G); end if;
       HD   := -P2(FD,G);                              -- fdot
       FDP  := Wi(DFdot(H0,HPG)+D2F(Alpha,HPG,HD));
       if Nd=1 and Np=1 then return P1(FDP,G); end if;
       FPPP := Wi(Scal(3)*D2F(Alpha,HPG,-P2(FPP,G)));
       if Nd=0 and Np=3 then return P1(FPPP,G); end if;
       Message("Ks.Gder",Not_Implemented);
       return ScalZero;
     end;
   end Gder;

   procedure USmall(F: in Rep; H: in Q; R: in out Radius; Iter: in Positive := 10) is
      Rsave: constant Rep := R;
      M: constant Matrix(PartIndex,PartIndex) := DFmI(Scal(F),H);
      Eps: constant Scalar := Norm(H-Contr(Scal(F),M,H));
      MinR: constant Rep := Sup(Eps);
      MaxR: constant Rep := Rep(8)*MinR;
      DR: constant Rep := (MaxR-MinR)/Rep(Iter+1);
      K: Scalar;
   begin
      if R<MinR or R>MaxR then R :=  MinR+DR; end if;
      for Count in 1 .. Iter loop
         K := DCNorm(Scal(F),M,Ball(H,R));
         if Inf(R*(ScalOne-K)-Eps)>Zero then return; end if;
         R := MinR+Rep(Count)*DR;
      end loop;
      R := Rsave;
      Show("Ks.USmall: giving up");
   end USmall;

--- for numerical experiments

   package Packd is new Packed (Number => Number, Size => Size, PartSize => PartSize);

   function DF(Alpha: Scalar; H: Q) return Matrix is
   begin
     if Trunc then return Basic.DF(Alpha,H); end if;
     declare
         A: Vector(FullIndex);
      begin
         Pack(H,A);
         return Packd.DF(Approx(Alpha),Par(H),Rho(H),A);
      end;
   end DF;

   function DFmI(Alpha: Scalar; H: Q) return Matrix is
   begin
      if Trunc then return Basic.DFmI(Alpha,H); end if;
      declare
         A: Vector(FullIndex);
      begin
         Pack(H,A);
         return Packd.DFmI(Approx(Alpha),Par(H),Rho(H),A);
      end;
   end DFmI;

   function DUtildemI(Alpha: Scalar; H,G: Q) return Matrix is
   begin
      if Trunc then return Basic.DUtildemI(Alpha,H,G); end if;
      declare
         A,B: Vector(FullIndex);
      begin
         Pack(H,A);
         Pack(G,B);
         return Packd.DUtildemI(Approx(Alpha),Par(H),Rho(H),A,B);
      end;
   end DUtildemI;

   function L(Alpha: Scalar; H: Q) return Matrix is
   begin
     if Trunc then return Basic.L(Alpha,H); end if;
     declare
         A: Vector(FullIndex);
     begin
       Pack(H,A);
       return Packd.L(Approx(Alpha),Par(H),Rho(H),A);
     end;
   end L;

   function Wi(Alpha: Scalar; H0,G: Q) return Matrix is
   begin
     if Trunc then return Basic.Wi(Alpha,H0,G); end if;
     declare
         B0,A: Vector(FullIndex);
     begin
       Pack(H0,B0);
       Pack(G,A);
       return Packd.Wi(Approx(Alpha),Par(H0),Rho(H0),B0,A);
     end;
   end Wi;

   procedure Newton(Alpha: in Scalar; Eps: in Radius; H: in out Q) is
      A: Vector(FullIndex);
   begin
      TraceEnter("Ks.Newton");
      Pack(H,A);
      Packd.Newton(Approx(Alpha),Eps,Par(H),Rho(H),A);
      UnPack(A,Par(H),Rho(H),H);
      TraceLeave;
   end Newton;

   procedure NewtonU(Alpha,Lambda: in Scalar; G: in Q; Eps: in Radius; H: in out Q) is
      A,B: Vector(FullIndex);
   begin
      TraceEnter("Ks.NewtonU");
      Pack(H,A);
      Pack(G,B);
      Packd.NewtonU(Approx(Alpha),Approx(Lambda),B,Eps,Par(H),Rho(H),A);
      UnPack(A,Par(H),Rho(H),H);
      TraceLeave;
   end NewtonU;

  function ApproxEigen(A: Matrix; R: Rep) return Vector is
     J: Integer;
     S: Rep := Rep'Last;
     W: Vector(A'Range(2));
     B,V: Matrix(A'Range(1),A'Range(2));
  begin
     B := A;
     Add(-R,B);

     Rounding.Relax_Checks;
     Eigen(Transpose(B)*B,W,V);
     Rounding.Restore_Checks;

     for I in W'Range loop
       if abs(W(I))<S then J := I; S := abs(W(I)); end if;
     end loop;
     return GetColumn(J,V);
   end ApproxEigen;

   function ApproxBifVec(Alpha: Scalar; H: Q) return Q is
      G: Q;
   begin
      UnPack(ApproxEigen(DF(Alpha,H),One),Par(H),Rho(H),G);
      return G;
   end ApproxBifVec;

   function ApproxEigenErr(A: Matrix; R: Rep) return Rep is
     S: Rep := Rep'Last;
     W: Vector(A'Range(2));
     B: Matrix(A'Range(1),A'Range(2));
   begin
     B := A;
     Add(-R,B);

     Rounding.Relax_Checks;
     Eigen(Transpose(B)*B,W);
     Rounding.Restore_Checks;

     for I in W'Range loop
       S := Rep'Min(S,abs(W(I)));
     end loop;
     return S;
   end ApproxEigenErr;

   function ApproxDist2Bif(Alpha: Scalar; H: Q) return Radius is
   begin
     return ApproxEigenErr(DF(Alpha,H),One);
   end ApproxDist2Bif;

   procedure Focus(D: in Rep; Err,Err2: in Radius; Alpha: in out Scalar; H: in out Q) is
      Nmax: constant Integer := 100;
      Fac: constant Rep := -0.8;
      DD: Scalar := Scal(D);
      E1,E2: Radius;
      H1: Q := H;
   begin
      if not Trunc then
         Message("Ks.Focus",Numeric_Only);
      end if;
      E1 := ApproxDist2Bif(Alpha,H);
      for N in 0 .. Nmax loop
         Alpha := Alpha+DD;
         H := H1;
         Newton(Alpha,Err,H);
         E2 := ApproxDist2Bif(Alpha,H);
         if E2 < Err2 then return; end if;
         if E2 > E1 then DD := Fac*DD; end if;
         E1 := E2;
      end loop;
      Show("Ks.Focus: giving up. Error: ",E2);
   end Focus;

   procedure BigR(F0,DF,RS0: in Rep; RL: in out Rep; H0: in Q) is
      MaxIter:  constant Positive := 50;
      MaxK:     constant Radius   := 0.8;
      A:        constant Radius   := 1.5;
      B:        constant Radius   := 3.0;
      OnePlus:  constant Radius   := 1.2;
      OneMinus: constant Radius   := 0.8;
      M: constant Matrix(PartIndex,PartIndex) := DFmI(Scal(F0,F0+DF),H0);
      KM: constant Scalar := DCNorm(Scal(F0,F0+DF),M,H0);
      KL: Scalar := DCNorm(Scal(F0,F0+DF),M,Ball(H0,RL));
   begin
      while Sup(KL) > MaxK loop
         RL := OneMinus*RL;
         KL := DCNorm(Scal(F0,F0+DF),M,Ball(H0,RL));
         if RL<RS0 then Message("Ks.BigR 1",Giving_Up); end if;
      end loop;
      for I in 1 .. MaxIter loop
        if One/(One-Approx(KL)) < B/(One-Approx(KM)) then
          if A/(One-Approx(KM)) < One/(One-Approx(KL)) then return; end if;
          RL := OnePlus*RL;
        else
          RL := OneMinus*RL;
        end if;
        KL := DCNorm(Scal(F0,F0+DF),M,Ball(H0,RL));
      end loop;
      Message("Ks.BigR 2",Giving_Up);
   end BigR;

   procedure SetInterval(F0,RS0,Err: in Rep; DF,RL: in out Rep; H0: in Q) is
      MaxIter:  constant Positive := 100;
      DFmin:    constant Radius   := 1.0E-20;
      MaxK:     constant Radius   := 0.1;
      OneMinus: constant Radius   := 0.8;
      KM: Scalar;
      H: Q := H0;
   begin
      Newton(Scal(F0,F0+DF),Err,H);
      KM := DCNorm(Scal(F0,F0+DF),DFmI(Scal(F0,F0+DF),H),H);
      while Sup(KM) > MaxK loop
         H := H0;
         DF := Half*DF;
         Newton(Scal(F0,F0+DF),Err,H);
         KM := DCNorm(Scal(F0,F0+DF),DFmI(Scal(F0,F0+DF),H),H);
         if DF<DFmin then Message("Ks.SetInterval 1",Giving_Up); end if;
      end loop;
      while DF > DFmin loop
         H := H0;
         Newton(Scal(F0,F0+DF),Err,H);
         BigR(F0,DF,RS0,RL,H);
         if U(Scal(F0,F0+DF),H,RL) then return; end if;
         DF := OneMinus*DF;
         BigR(F0,DF,RS0,RL,H);
      end loop;
      Message("Ks.SetInterval 2",Giving_Up);
   end SetInterval;

end Ks;
