with Ada.Text_IO, Reps, Linear, Fouriers_Init, Basics;
use  Ada.Text_IO, Reps, Linear, Fouriers_Init;

generic

  Number,Size,PartSize: in Positive;

  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Scal(R1,R2: Rep) return Scalar is <>;
  with function Ball(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function Approx(S: Scalar) return Rep is <>;
  with function Inf(S: Scalar) return Rep is <>;
  with function Sup(S: Scalar) return Rep is <>;
  with function InfAbs(S: Scalar) return Rep is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Envelope(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with function "*"(R: Rep; S: Scalar) return Scalar is <>;
  with function "/"(S: Scalar; R: Rep) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Widen(S: Scalar) return Scalar is <>;
  with function Piece(M,N: Positive; S: Scalar) return Scalar is <>;
  with procedure Root4(C0,C1,C3: in Scalar; S0: in Radius; S: out Scalar; Exists: out Boolean) is <>;
  with procedure Show(S: in Scalar; NewLine: in Boolean := True) is <>;
  with procedure Get(F: in File_Type; S: out Scalar; Decimal: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := True) is <>;

package Ks is

   package Basic is new Basics (Number => Number, Size => Size, PartSize => PartSize, Scalar => Scalar);
   package Qf renames Basic.Qb;
   package Qf_Ops renames Basic.Qb_Ops;
   package Qf_IO renames Basic.Qb_IO;
   use Qf,Qf_IO;

   package FSLinear renames Qf.QSLinear;
   package FSLinear_Ops renames Qf_Ops.QSLinear_Ops;
   package FSLinear_IO renames Qf_IO.QSLinear_IO;
   use FSLinear, FSLinear_Ops, FSLinear_IO;

   subtype Freq is Basic.Freq;
   subtype FullIndex is Basic.FullIndex;       -- vector index
   subtype PartIndex is Basic.PartIndex;       -- vector index

   subtype PRVector is Vector(PartIndex);
   subtype PRMatrix is Matrix(PartIndex,PartIndex);
   subtype PSVector is SVector(PartIndex);
   subtype PSMatrix is SMatrix(PartIndex,PartIndex);

   type BlockOp is private;

   procedure ShowParam;                                           -- show some parameters

   function F(Alpha: Scalar; H: Q) return Q renames Basic.F;      -- F(H) := (alpha/4)(-D^(-2)H +D^(-3) H^2)
   function DF(Alpha: Scalar; H,G: Q) return Q renames Basic.DF;  -- DF(H)G

   function Utilde(Alpha,Lambda: Scalar; H,G: Q) return Q renames Basic.Utilde;
   function DUtilde(Alpha: Scalar; H0,H,G: Q) return Q renames Basic.DUtilde;

   function FF(Alpha: Scalar; H: Q) return Q;                           -- F(H)-H
   function DFF(Alpha: Scalar; H,G: Q) return Q;                        -- DFF(H)G
   function Fdot(H: Q) return Q;                                        -- derivative of F w.r.to alpha
   function DFdot(H,G: Q) return Q;                                     -- derivative of Fdot
   function D2F(Alpha: Scalar; G1,G2: Q) return Q;                      -- 2nd derivative of F

   function Contr(Alpha: Scalar; M: Matrix; H: Q) return Q;             -- F(H)-M(F(H)-H)
   function DContr(Alpha: Scalar; M: Matrix; H,G: Q) return Q;          -- derivative of Contr
   function DCNorm(Alpha: Scalar; M: Matrix; H: Q) return Scalar;       -- norm of DContr
   function UContr(Alpha,Lambda: Scalar; M: Matrix; H,G: Q) return Q;
   function DUContr(Alpha: Scalar; M: Matrix; H0,H,G: Q) return Q;
   function UNorm(Alpha: Scalar; M: Matrix; H,G: Q) return Scalar;

   procedure Convert(Ar,Ai: in Vector; A: out BlockOp);
   function Norm(A,B: BlockOp; R: Radius) return Scalar;
   function Inv(A: BlockOp) return BlockOp;                             -- 1/A
   function "*"(A,B: BlockOp) return BlockOp;                           -- A*B
   function "*"(A: BlockOp; G: Q) return Q;                             -- A*G

   function T(Alpha: Scalar; G: Q) return Q;                            -- -(4*D^4+alpha*D^2)
   function Tinv(Alpha: Scalar; G: Q) return Q;                         -- T^(-1)
   function DTinv(Alpha: Scalar; G: Q) return Q;                        -- D(T^(-1))
   function LmT(Alpha: Scalar; H,DH,G,DG: Q) return Q;                  -- 2*alpha*(G*DH+H*DG)
   function L(Alpha: Scalar; H,DH,G: Q) return Q;                       -- LmT + T
   procedure MakeLinear(Alpha: in Scalar; H: in Q; SSi,TT: out BlockOp; -- define linear operators
                EV: out Matrix; EVinv: out SMatrix);
   function ASi(Alpha: Scalar; EVi: SMatrix; H,DH: Q; EV: Matrix;       -- E^(-1)*LmT*S^(-1)*E
                TT,SSi: BlockOp; G: Q) return Q;
   function ASiNorm(Alpha: Scalar; EVi: SMatrix; H,DH: Q; EV: Matrix;   -- norm of ASi
                TT,SSi: BlockOp) return Scalar;

   function W(Alpha: Scalar; H0,H,G: Q) return Q renames Basic.W;       -- (T-P2*T*P1)H , where T=DFF(H0)+c*P1
   function W1(Alpha: Scalar; H: Q) return Q renames Basic.W1;
   function W2(Alpha: Scalar; H0,H,G: Q) return Q renames Basic.W2;

   function Wi1(Alpha: Scalar; H: Q) return Q;                          -- W1^(-1) H
   function Wi3(Alpha: Scalar; Wi0: Matrix; H: Q) return Q;             -- (Wi0+Wi1)H
   function WiW(Alpha: Scalar; Wi0: Matrix; H0,H,G: Q) return Q;        -- [I-Wi3*W]H
   function WiW(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q;      -- WiW when A is known
   function WiWNorm(Alpha: Scalar; Wi0: Matrix; H0,G: Q) return Scalar; -- norm of WiW
   function Wi(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q;       -- W^(-1)H
   function PWiP(Alpha,A: Scalar; Wi0: Matrix; H0,H,G: Q) return Q;     -- P2 W^(-1) P2 H
   function UT(Alpha,Lambda: Scalar; H,G: Q; R: Radius;
               NA,NL: Positive := 1) return Boolean;                    -- checks solution of P2(FF(H0))=0
   function U(Alpha: Scalar; H: Q; R: Radius) return Boolean;           -- checks if F(H0)=H0
   function GG(Alpha: Scalar; H,G: Q) return Scalar;                    -- P1(FF(H))
   function Gder(Nd,Np: Integer; Alpha,A: Scalar;
                 Wi0: Matrix; H0,G: Q) return Scalar;                   -- (alpha,lambda) derivatives of GG
   procedure USmall(F: in Rep; H: in Q; R: in out Radius; Iter: in Positive := 10);    -- see paper

--- for numerical experiments

   function DF(Alpha: Scalar; H: Q) return Matrix;              -- approx DF(H)
   function DFmI(Alpha: Scalar; H: Q) return Matrix;            -- approx (DF(H)-I)^(-1)+I
   function DUtildemI(Alpha: Scalar; H,G: Q) return Matrix;
   function L(Alpha: Scalar; H: Q) return Matrix;               -- approx -(4*D^4+alpha*D^2)G+2*alpha*D(G*H)
   function Wi(Alpha: Scalar; H0,G: Q) return Matrix;           -- approx 1/(T-P2*T*P1) , where T=DFF(H0)+c*P1

   procedure Newton(Alpha: in Scalar; Eps: in Radius; H: in out Q);                    -- approx Newton
   procedure NewtonU(Alpha,Lambda: in Scalar; G: in Q; Eps: in Radius; H: in out Q);   -- approx Newton
   function ApproxBifVec(Alpha: Scalar; H: Q) return Q;                                -- approx bifurcation direction
   procedure Focus(D: in Rep; Err,Err2: in Radius; Alpha: in out Scalar; H: in out Q); -- find approx bifurcation point
   procedure BigR(F0,DF,RS0: in Rep; RL: in out Rep; H0: in Q);                        -- see paper
   procedure SetInterval(F0,RS0,Err: in Rep; DF,Rl: in out Rep; H0: in Q);             -- see paper

private

   type BlockOp is
     record
       Re: PSVector;
       Im: PSVector;
     end record;

   function BlockSize(K: Positive; A: BlockOp; Check: Boolean := False) return Integer;
--   returns 1 if the Kth eigenvalue is real, 2 if it is complex, 0 if Check = True and something is wrong
   function InvTminusImagNorm(TT: BlockOp; K: Integer; R: Radius) return Scalar;
   function ApproxEigen(A: Matrix; R: Rep) return Vector;
--   computes approximate eigenvector for matrix A and eigenvalue R
   function ApproxEigenErr(A: Matrix; R: Rep) return Rep;
--   returns error for eigenvector for matrix A and eigenvalue R
   function ApproxDist2Bif(Alpha: Scalar; H: Q) return Radius;
--   indicator of how close (Alpha,H) is to a bifurcation point.

end Ks;
