with Ada.Text_IO, Messages, Reps.Ops;
use Ada.Text_IO, Messages, Reps.Ops;

pragma Elaborate_All (Ada.Text_IO, Reps.Ops);

package body Linear.Ops is

  function CountLarge(B: Vector; Cut: Rep := Zero) return Integer is
    N: Integer := 0;
  begin
    for I in B'Range loop
      if B(I)>Cut then N := N+1; end if;
    end loop;
    return N;
  end CountLarge;

  procedure Add(B1: in Vector; B2: in out Vector) is
  begin
    for I in B2'Range loop
      B2(I) := B2(I)+B1(I);
    end loop;
  end Add;

  procedure Add(A1: in Matrix; A2: in out Matrix) is
  begin
    for I in A2'Range(1) loop
      for J in A2'Range(2) loop
         A2(I,J) := A2(I,J)+A1(I,J);
      end loop;
    end loop;
  end Add;

  procedure Add(R: in Rep; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      A(I,I) := A(I,I)+R;
    end loop;
  end Add;

  procedure Sub(B1: in Vector; B2: in out Vector) is
  begin
    for I in B2'Range loop
      B2(I) := B2(I)-B1(I);
    end loop;
  end Sub;

  procedure Sub(A1: in Matrix; A2: in out Matrix) is
  begin
    for I in A2'Range(1) loop
      for J in A2'Range(2) loop
         A2(I,J) := A2(I,J)-A1(I,J);
      end loop;
    end loop;
  end Sub;

  procedure Mult(R: in Rep; B: in out Vector) is
  begin
    for I in B'Range loop
      B(I) := R*B(I);
    end loop;
  end Mult;

  procedure Mult(R: in Rep; A: in out Matrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        A(I,J) := R*A(I,J);
      end loop;
    end loop;
  end Mult;

  procedure Mult(A1: in Matrix; A2: in out Matrix) is
    R: Rep;
    B: Vector(A2'Range(1));
  begin
    for J in B'Range loop
      for K in B'Range loop
        B(K) := A2(K,J);
      end loop;
      for I in B'Range loop
        R := Zero;
        for K in B'Range loop
          R := R+A1(I,K)*B(K);
        end loop;
        A2(I,J) := R;
      end loop;
    end loop;
  end Mult;

  procedure DMult(Bf: in Vector; B: in out Vector) is
  begin
    for I in B'Range loop
      B(I) := B(I)*Bf(I);
    end loop;
  end DMult;

  function Transpose(A: Matrix) return Matrix is
     B: Matrix(A'Range(2),A'Range(1));
     begin
        for I in A'Range(1) loop
           for j in A'Range(2) loop
              B(J,I) := A(I,J);
           end loop;
        end loop;
        return B;
     end Transpose;

  function "-"(B: Vector) return Vector is
     Bt: Vector(B'Range);
  begin
    for I in B'Range loop
      Bt(I) := -B(I);
    end loop;
    return Bt;
   end "-";

  function "+"(B1,B2: Vector) return Vector is
    B: Vector(B2'Range);
  begin
    for I in B'Range loop
      B(I) := B2(I)+B1(I);
    end loop;
    return B;
  end "+";

  function "+"(A1,A2: in Matrix) return Matrix is
    A: Matrix(A2'Range(1),A2'Range(2));
  begin
    for I in A'Range loop
      for J in A'Range loop
         A(I,J) := A2(I,J)+A1(I,J);
      end loop;
    end loop;
    return A;
  end "+";

  function "-"(B1,B2: Vector) return Vector is
    B: Vector(B2'Range);
  begin
    for I in B'Range loop
      B(I) := B2(I)-B1(I);
    end loop;
    return B;
  end "-";

  function "-"(A1,A2: in Matrix) return Matrix is
    A: Matrix(A2'Range(1),A2'Range(2));
  begin
    for I in A'Range loop
      for J in A'Range loop
         A(I,J) := A2(I,J)-A1(I,J);
      end loop;
    end loop;
    return A;
  end "-";

  function "*"(B1,B2: Vector) return Rep is
    R: Rep;
  begin
    R := Zero;
    for I in B2'Range loop
      R := R+B1(I)*B2(I);
    end loop;
    return R;
  end "*";

  function "*"(R: Rep; B: Vector) return Vector is
    Br: Vector(B'Range);
  begin
    for I in B'Range loop
      Br(I) := R*B(I);
    end loop;
    return Br;
  end "*";

  function "*"(R: Rep; A: Matrix) return Matrix is
    Ar: Matrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Ar(I,J) := R*A(I,J);
      end loop;
    end loop;
    return Ar;
  end "*";

  function "*"(B1,B2: Vector) return Vector is
    B: Vector(B2'Range);
  begin
    for I in B2'Range loop
      B(I) := B1(I)*B2(I);
    end loop;
    return B;
  end "*";

  function "*"(A: Matrix; B: Vector) return Vector is
    R: Rep;
    Br: Vector(A'Range(1));
  begin
    for I in Br'Range loop
      R := Zero;
      for J in B'Range loop
        R := R+A(I,J)*B(J);
      end loop;
      Br(I) := R;
    end loop;
    return Br;
  end "*";

  function "*"(B: Vector; A: Matrix) return Vector is
    R: Rep;
    Br: Vector(A'Range(2));
  begin
    for I in Br'Range loop
      R := Zero;
      for J in B'Range loop
        R := R+B(J)*A(J,I);
      end loop;
      Br(I) := R;
    end loop;
    return Br;
  end "*";

  function "*"(A1,A2: Matrix) return Matrix is
    R: Rep;
    A: Matrix(A1'Range(1),A2'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        R := Zero;
        for K in A2'Range(1) loop
          R := R+A1(I,K)*A2(K,J);
        end loop;
        A(I,J) := R;
      end loop;
    end loop;
    return A;
  end "*";

--- linpack inverse* for real matrices ----------------------------------------------

   procedure dgefa(n: in Integer; a: in out Matrix; ipvt: out Pivot; info: out Integer) is
-- translated from linpack routine dgefa, http://www.netlib.org/linpack/
      l: Integer;
      t: Rep;

   begin
      info := 0;
      for k in 1 .. n-1 loop

--        find l := pivot index;

         t := abs(a(n,k));
         l := n;
         for m in reverse k .. n-1 loop
            if (abs(a(m,k)) >= t) then
               t := abs(a(m,k));
               l := m;
            end if;
         end loop;
         ipvt(k) := l;

--        zero pivot implies this column already triangularized

         if (a(l,k) = Zero) then
            info := k;
         else

--           interchange if necessary

            if (l /= k) then
               t := a(l,k);
               a(l,k) := a(k,k);
               a(k,k) := t;
            end if;

--           compute multipliers

            t := -One/a(k,k);
            for m in k+1 .. n loop
               a(m,k) := t*a(m,k);
            end loop;

--           row elimination with column indexing

            for j in k+1 .. n loop
               t := a(l,j);
               if (l /= k) then
                  a(l,j) := a(k,j);
                  a(k,j) := t;
               end if;
               for m in k+1 .. n loop
                  a(m,j) := a(m,j) + t*a(m,k);
               end loop;
            end loop;
         end if;
      end loop;

      ipvt(n) := n;
      if (a(n,n) = Zero) then info := n; end if;
   end dgefa;

   procedure dgesl(n,job: in Integer; ipvt: in Pivot; a: in Matrix; b: in out Vector) is
-- translated from linpack routine dgesl, http://www.netlib.org/linpack/
      l: Integer;
      t: Rep;

   begin
      if (job = 0) then

--        job = 0 , solve  a * x := b;
--        first solve  l*y := b;

         for k in 1 .. n-1 loop
            l := ipvt(k);
            t := b(l);
            if (l /= k) then
               b(l) := b(k);
               b(k) := t;
            end if;
            for m in k+1 .. n loop
               b(m) := b(m) + t*a(m,k);
            end loop;
         end loop;

--        now solve  u*x := y;

         for k in reverse 1 .. n loop
            b(k) := b(k)/a(k,k);
            t := -b(k);
            for m in 1 .. k-1 loop
               b(m) := b(m) + t*a(m,k);
            end loop;
         end loop;
         return;
      end if;

--     job = nonzero, solve  trans(a) * x := b;
--     first solve  trans(u)*y := b;

      for k in 1 .. n loop
         t := Zero;
         for m in 1 .. k-1 loop
            t := t + a(m,k)*b(m);
         end loop;
         b(k) := (b(k) - t)/a(k,k);
      end loop;

--     now solve trans(l)*x := y;

      for k in reverse 1 .. n-1 loop
         t := b(k);
         for m in k+1 .. n loop
            t := t + a(m,k)*b(m);
         end loop;
         b(k) := t;
         l := ipvt(k);
         if (l /= k) then
            t := b(l);
            b(l) := b(k);
            b(k) := t;
         end if;
      end loop;
   end dgesl;

   procedure dgedi(n,job: in Integer; ipvt: in Pivot; a: in out Matrix; work: in out Vector; det1,det2: out Rep) is
-- translated from linpack routine dgedi, http://www.netlib.org/linpack/
      l: integer;
      ten: constant Rep := 1.0E+1;
      t: Rep;

   begin
--     compute determinant

      if (job /= 1) then
         det1 := One;
         det2 := Zero;
         for i in 1 .. n loop
            if (ipvt(i) /= i) then det1 := -det1; end if;
            det1 := a(i,i)*det1;
            exit when (det1 = Zero);
            while (abs(det1) < One) loop
               det1 := ten*det1;
               det2 := det2 - One;
            end loop;
            while (abs(det1) >= ten) loop
               det1 := det1/ten;
               det2 := det2 + One;
            end loop;
         end loop;
      end if;

--     compute inverse(u)

      if (job = 10) then return; end if;

      for k in 1 .. n loop
         a(k,k) := One/a(k,k);
         t := -a(k,k);
         for m in 1 .. k-1 loop
            a(m,k) := t*a(m,k);
         end loop;
         for j in k+1 .. n loop
            t := a(k,j);
            a(k,j) := Zero;
            for m in 1 .. k loop
               a(m,j) := a(m,j) + t*a(m,k);
            end loop;
         end loop;
      end loop;

--     form inverse(u)*inverse(l)

      for k in reverse 1 .. n-1 loop
         for i in k+1 .. n loop
            work(i) := a(i,k);
            a(i,k) := Zero;
         end loop;
         for j in k+1 .. n loop
            t := work(j);
            for m in 1 .. n loop
               a(m,k) := a(m,k) + t*a(m,j);
            end loop;
         end loop;
         l := ipvt(k);
         if (l /= k) then
            for m in 1 .. n loop
               t := a(m,k);
               a(m,k) := a(m,l);
               a(m,l) := t;
            end loop;
         end if;
      end loop;
   end dgedi;

  procedure Solve(A: in Matrix; B: in Vector; X: out Vector) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Ipvt: Pivot(1 .. N);
    Ar: Matrix(1 .. N,1 .. N);
  begin
    Ar := A;
    X := B;
    Dgefa(N,Ar,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgesl(N,0,Ipvt,Ar,X);
  end Solve;

  procedure TSolve(A: in Matrix; B: in Vector; X: out Vector) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Ipvt: Pivot(1 .. N);
    Ar: Matrix(1 .. N,1 .. N);
  begin
    Ar := A;
    X := B;
    Dgefa(N,Ar,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgesl(N,1,Ipvt,Ar,X);
  end TSolve;

  function Det(A: Matrix) return Rep is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Det1,Det2: Rep;
    Ipvt: Pivot(1 .. N);
    Work: Vector(1 .. N);
    Ar: Matrix(1 .. N,1 .. N);
  begin
    Ar := A;
    Dgefa(N,Ar,Ipvt,Info);
    Dgedi(N,10,Ipvt,Ar,Work,Det1,Det2);
    return Det1*Exp(Log(1.0E+1)*Det2);
  end Det;

  procedure Invert(A: in out Matrix) is
    pragma Suppress(Storage_Check);
    N: constant Integer := A'Last(1);
    Info: Integer;
    Det1,Det2: Rep;
    Ipvt: Pivot(1 .. N);
    Work: Vector(1 .. N);
  begin
    Dgefa(N,A,Ipvt,Info);
    if Info>0 then
      Show("Warning: Problem with diagonal ",Info);
    end if;
    Dgedi(N,1,Ipvt,A,Work,Det1,Det2);
  end Invert;

--- eispack eigen* for real symmetric matrices --------------------------------------

   function dsign(a,b: Rep) return Rep is
---  ancient fortran routine
   begin
      if B<Zero then return -abs(A); end if;
      return abs(A);
   end dsign;

   procedure tred1(n: in Integer; a: in out Matrix; d,e,e2: out Vector) is
-- translated from eispack routine tred1, http://www.netlib.org/eispack/
      l: Integer;
      f,g,h,scale: Rep;

begin

      for i in 1 .. n loop
         d(i) := a(n,i);
         a(n,i) := a(i,i);
      end loop;

      for i in reverse 1 .. n loop
         l := i - 1;
         h := Zero;
         scale := Zero;

         if (l < 1) then
            e(i) := Zero;
            e2(i) := Zero;
            goto L300;
         else

--     .......... scale row (algol tol then not needed) ..........

            for k in 1 .. l loop
               scale := scale + abs(d(k));
            end loop;
            if (scale = Zero) then
               for j in 1 .. l loop
                  d(j) := a(l,j);
                  a(l,j) := a(i,j);
                  a(i,j) := Zero;
               end loop;
               e(i) := Zero;
               e2(i) := Zero;
               goto L300;
            end if;
         end if;

         for k in 1 .. l loop
            d(k) := d(k) / scale;
            h := h + d(k) * d(k);
         end loop;
--
         e2(i) := scale * scale * h;
         f := d(l);
         g := -dsign(sqrt(h),f);
         e(i) := scale * g;
         h := h - f * g;
         d(l) := f - g;
         if (l /= 1) then

--     .......... form a*u ..........

            for j in 1 .. l loop
               e(j) := Zero;
            end loop;

            for j in 1 .. l loop
               f := d(j);
               g := e(j) + a(j,j) * f;
               if (l >= j+1) then
                  for k in j+1 .. l loop
                     g := g + a(k,j) * d(k);
                     e(k) := e(k) + a(k,j) * f;
                  end loop;
               end if;
               e(j) := g;
            end loop;

--     .......... form p ..........

            f := Zero;
            for j in 1 .. l loop
               e(j) := e(j) / h;
               f := f + e(j) * d(j);
            end loop;
            h := f / (h + h);

--     .......... form q ..........

            for j in 1 .. l loop
               e(j) := e(j) - h * d(j);
            end loop;

--     .......... form reduced a ..........

            for j in 1 .. l loop
               f := d(j);
               g := e(j);
               for k in j .. l loop
                  a(k,j) := a(k,j) - f * e(k) - g * d(k);
               end loop;
            end loop;
         end if;

         for j in 1 .. l loop
            f := d(j);
            d(j) := a(l,j);
            a(l,j) := a(i,j);
            a(i,j) := f * scale;
         end loop;

  <<L300>> null;

      end loop;
   end tred1;

   procedure tred2(n: in Integer; a: in Matrix; d,e: out Vector; z: out Matrix) is
-- translated from eispack routine tql1, http://www.netlib.org/eispack/
      l: integer;
      f,g,h,hh,scale: Rep;

   begin
      if (n = 1) then
         d(1) := a(1,1);
         z(1,1) := One;
         e(1) := Zero;
         return;
      end if;

      for i in 1 .. n loop
         for j in i .. n loop
            z(j,i) := a(j,i);
         end loop;
         d(i) := a(n,i);
      end loop;

      for i in reverse 2 .. n loop
         l := i - 1;
         h := Zero;
         scale := Zero;

         if (l < 2) then
            e(i) := d(l);
            for j in 1 .. l loop
               d(j) := z(l,j);
               z(i,j) := Zero;
               z(j,i) := Zero;
            end loop;
            goto L290;
         else

--     .......... scale row (algol tol then not needed) ..........

            for k in 1 .. l loop
               scale := scale + abs(d(k));
            end loop;
            if (scale = Zero) then
               e(i) := d(l);
               for j in 1 .. l loop
                  d(j) := z(l,j);
                  z(i,j) := Zero;
                  z(j,i) := Zero;
               end loop;
               goto L290;
            end if;
         end if;

         for k in 1 .. l loop
            d(k) := d(k) / scale;
            h := h + d(k) * d(k);
         end loop;

         f := d(l);
         g := -dsign(sqrt(h),f);
         e(i) := scale * g;
         h := h - f * g;
         d(l) := f - g;

--     .......... form a*u ..........

         for j in 1 .. l loop
            e(j) := Zero;
         end loop;

         for j in 1 .. l loop
            f := d(j);
            z(j,i) := f;
            g := e(j) + z(j,j) * f;
            if (l > j) then
               for k in j+1 .. l loop
                  g := g + z(k,j) * d(k);
                  e(k) := e(k) + z(k,j) * f;
               end loop;
            end if;
            e(j) := g;
         end loop;

--     .......... form p ..........

         f := Zero;

         for j in 1 .. l loop
            e(j) := e(j) / h;
            f := f + e(j) * d(j);
         end loop;
         hh := f / (h + h);

--     .......... form q ..........

         for j in 1 .. l loop
            e(j) := e(j) - hh * d(j);
         end loop;

--     .......... form reduced a ..........

         for j in 1 .. l loop
            f := d(j);
            g := e(j);
            for k in j .. l loop
               z(k,j) := z(k,j) - f * e(k) - g * d(k);
            end loop;
            d(j) := z(l,j);
            z(i,j) := Zero;
         end loop;

  <<L290>> null;

         d(i) := h;
      end loop;

--     .......... accumulation of transformation matrices ..........

      for i in 2 .. n loop
         l := i - 1;
         z(n,l) := z(l,l);
         z(l,l) := One;
         h := d(i);
          if (h /= Zero) then
            for k in 1 .. l loop
               d(k) := z(k,i) / h;
            end loop;

            for j in 1 .. l loop
               g := Zero;
               for k in 1 .. l loop
                  g := g + z(k,i) * z(k,j);
               end loop;
               for k in 1 .. l loop
                  z(k,j) := z(k,j) - g * d(k);
               end loop;
            end loop;
         end if;
         for k in 1 .. l loop
            z(k,i) := Zero;
         end loop;
      end loop;

      for i in 1 .. n loop
         d(i) := z(n,i);
         z(n,i) := Zero;
      end loop;
      z(n,n) := One;
      e(1) := Zero;
   end tred2;

   procedure tql1(n: in Integer; d,e: in out Vector; ierr: out Integer) is
-- translated from eispack routine tql1, http://www.netlib.org/eispack/
     j,m,l1,l2: integer;
     c,c2,c3,dl1,el1,f,g,h,p,r,s,s2,tst1,tst2: Rep;

   begin
      ierr := 0;
      if (n = 1) then return; end if;
      for i in 2 .. n loop
          e(i-1) := e(i);
      end loop;
      f := Zero;
      tst1 := Zero;
      e(n) := Zero;

      for l in 1 .. n loop
         j := 0;
         h := abs(d(l)) + abs(e(l));
         if (tst1 < h) then tst1 := h; end if;

--     .......... look for small sub-diagonal element ..........

         m := l;
         loop
            tst2 := tst1 + abs(e(m));
            exit when (tst2 = tst1);
            m := m+1;
         end loop;

         if (m /= l) then
            loop
               if (j = 30) then

--     .......... set error -- no convergence to an
--                eigenvalue after 30 iterations ..........

                  ierr := l;
                  return;
               end if;

               j := j + 1;

--     .......... form shift ..........

               l1 := l + 1;
               l2 := l1 + 1;
               g := d(l);
               p := (d(l1) - g) / (Two * e(l));
               r := sqrt(One+p*p);
               d(l) := e(l) / (p + dsign(r,p));
               d(l1) := e(l) * (p + dsign(r,p));
               dl1 := d(l1);
               h := g - d(l);
               for i in l2 .. n loop
                  d(i) := d(i) - h;
               end loop;
               f := f + h;

--     .......... ql transformation ..........

               p := d(m);
               c := One;
               c2 := c;
               el1 := e(l1);
               s := Zero;
               for i in reverse l .. m-1 loop
                  c3 := c2;
                  c2 := c;
                  s2 := s;
                  g := c * e(i);
                  h := c * p;
                  r := sqrt(p*p+e(i)*e(i));
                  e(i+1) := s * r;
                  s := e(i) / r;
                  c := p / r;
                  p := c * d(i) - s * g;
                  d(i+1) := h + s * (c * g + s * d(i));
               end loop;
--
               p := -s * s2 * c3 * el1 * e(l) / dl1;
               e(l) := s * p;
               d(l) := c * p;
               tst2 := tst1 + abs(e(l));
               exit when (tst2 = tst1);
            end loop;
         end if;

         p := d(l) + f;

--     .......... order eigenvalues ..........

         if (l = 1) then d(1) := p; end if;

         for i in reverse 2 .. l loop
            if (p >= d(i-1)) then
               d(i) := p;
               exit;
             end if;
            d(i) := d(i-1);
            if (i = 2) then d(1) := p; end if;
         end loop;
      end loop;
   end tql1;

   procedure tql2(n: in Integer; d,e: in out Vector; z: in out Matrix; ierr: out Integer) is
-- translated from eispack routine tql2, http://www.netlib.org/eispack/
      j,m,l1,l2: Integer;
      c,c2,c3,dl1,el1,f,g,h,p,r,s,s2,tst1,tst2: Rep;

   begin
      ierr := 0;
      if (n = 1) then return; end if;
      for i in 2 .. n loop
         e(i-1) := e(i);
      end loop;
      f := Zero;
      tst1 := Zero;
      e(n) := Zero;

      for l in 1 .. n loop
         j := 0;
         h := abs(d(l)) + abs(e(l));
         if (tst1 < h) then tst1 := h; end if;

--     .......... look for small sub-diagonal element ..........

         m := l;
         loop
            tst2 := tst1 + abs(e(m));
            exit when (tst2 = tst1);
            m := m+1;
         end loop;

         if (m /= l) then
            loop
               if (j = 30) then

--     .......... set error -- no convergence to an
--                eigenvalue after 30 iterations ..........

                  ierr := l;
                  return;
               end if;

               j := j + 1;

--     .......... form shift ..........

               l1 := l + 1;
               l2 := l1 + 1;
               g := d(l);
               p := (d(l1) - g) / (Two * e(l));
               r := sqrt(One+p*p);
               d(l) := e(l) / (p + dsign(r,p));
               d(l1) := e(l) * (p + dsign(r,p));
               dl1 := d(l1);
               h := g - d(l);
               for i in l2 .. n loop
                  d(i) := d(i) - h;
               end loop;
               f := f + h;

--     .......... ql transformation ..........

               p := d(m);
               c := One;
               c2 := c;
               el1 := e(l1);
               s := Zero;
               for i in reverse l .. m-1 loop
                  c3 := c2;
                  c2 := c;
                  s2 := s;
                  g := c * e(i);
                  h := c * p;
                  r := sqrt(p*p+e(i)*e(i));
                  e(i+1) := s * r;
                  s := e(i) / r;
                  c := p / r;
                  p := c * d(i) - s * g;
                  d(i+1) := h + s * (c * g + s * d(i));

--     .......... form vector ..........
                  for k in 1 .. n loop
                     h := z(k,i+1);
                     z(k,i+1) := s * z(k,i) + c * h;
                     z(k,i) := c * z(k,i) - s * h;
                  end loop;
               end loop;

               p := -s * s2 * c3 * el1 * e(l) / dl1;
               e(l) := s * p;
               d(l) := c * p;
               tst2 := tst1 + abs(e(l));
               exit when (tst2 = tst1);
            end loop;
         end if;

         d(l) := d(l) + f;
      end loop;

--     .......... order eigenvalues and eigenvectors ..........

      for i in 1 .. n-1 loop
         m := i;
         p := d(i);
         for l in i+1 .. n loop
            if (d(l) < p) then
               m := l;
               p := d(l);
            end if;
         end loop;
         if (m /= i) then
            d(m) := d(i);
            d(i) := p;
            for l in 1 .. n loop
               p := z(l,i);
               z(l,i) := z(l,m);
               z(l,m) := p;
            end loop;
         end if;
      end loop;
   end tql2;

   procedure SortEigen(W: in out Vector; V: in out Matrix) is
     X,Y: Integer;
     A: Pivot(W'Range);
     U: Vector(W'Range);

     function GT(I,J: Integer) return Boolean is
     begin
       return abs(W(I))>abs(W(J));
     end GT;

     procedure Qsort(L,R: Integer) is
       I,J: integer;
     begin
       I:=L; J:=R; X:=A((L+R)/2);
       loop
         while GT(A(I),X) loop I:=I+1; end loop;
         while GT(X,A(J)) loop J:=J-1; end loop;
         if I <= J then
           Y:=A(I); A(I):=A(J); A(J):=Y;
           I:=I+1; J:=J-1;
         end if;
         exit when I>J;
       end loop;
       if L<J then Qsort(L,J); end if;
       if I<R then Qsort(I,R); end if;
     end Qsort;

   begin
     for J in W'Range loop A(J) := J; end loop;
     Qsort(W'First,W'Last);

     U := W;
     for J in W'Range loop W(J) := U(A(J)); end loop;

     for I in W'Range loop
       for J in W'Range loop U(J) := V(I,J); end loop;
       for J in W'Range loop V(I,J) := U(A(J)); end loop;
     end loop;
   end SortEigen;

   procedure Eigen(A: in Matrix; W: out Vector) is
     pragma Suppress(Storage_Check);
     N: constant Integer := A'Last(1);
     Ierr: Integer;
     Fv1,Fv2: Vector(1 .. N);
     Ar: Matrix(1 .. N,1 .. N);
   begin
     Ar := A;
     Tred1(N,Ar,W,Fv1,Fv2);
     Tql1(N,W,Fv1,Ierr);
     if Ierr>0 then
       Show("Error: Problem with eigenvalue number ",Ierr);
     end if;
   end Eigen;

   procedure Eigen(A: in Matrix; W: out Vector; V: out Matrix) is
     pragma Suppress(Storage_Check);
     N: constant Integer := A'Last(1);
     Ierr: Integer;
     Fv1: Vector(1 .. N);
     Ar: Matrix(1 .. N,1 .. N);
   begin
     Ar := A;
     Tred2(N,Ar,W,Fv1,V);
     Tql2(N,W,Fv1,V,Ierr);
     if Ierr>0 then
       Show("Error: Problem with eigenvalue number ",Ierr);
     end if;
     SortEigen(W,V);
   end Eigen;

-- eispack eigen* for complex matrices --------------------------------------------

   procedure cdiv(ar,ai,br,bi: in Rep; cr,ci: out Rep) is
-- complex division, eispack style
      s: constant Rep := br*br+bi*bi;
   begin
      cr := (ar*br+ai*bi)/s;
      ci := (ai*br-ar*bi)/s;
   end cdiv;

   procedure csroot(xr,xi: in Rep; yr,yi: out Rep) is
-- complex square root, eispack style
-- branch chosen so that yr>=0 and sign(yi)=sign(xi)
      s: Rep;
   begin
      s := sqrt(Half*(sqrt(xr*xr+xi*xi)+abs(xr)));
      if xr >= Zero then yr := s; end if;
      if xi <  Zero then s := -s; end if;
      if xr <= Zero then yi := s; end if;
      if xr <  Zero then yr := Half*(xi/yi); end if;
      if xr >  Zero then yi := Half*(xi/yr); end if;
   end csroot;

   procedure cbal(n: in Integer; ar,ai: in out Matrix; low,igh: out Integer; scale: out Vector) is
-- translated from eispack routine cbal, http://www.netlib.org/eispack/
     allzero,noconv: Boolean;
     k,l,m: Integer;
     c,f,g,r,s,b2,radix: Rep;

--     .......... row and column exchange ..........
      procedure exc(m,j: in integer) is
      begin
         scale(m) := Rep(j);
         if (j = m) then return; end if;
         for i in 1 .. l loop
            f := ar(i,j);
            ar(i,j) := ar(i,m);
            ar(i,m) := f;
            f := ai(i,j);
            ai(i,j) := ai(i,m);
            ai(i,m) := f;
         end loop;
         for i in k .. n loop
            f := ar(j,i);
            ar(j,i) := ar(m,i);
            ar(m,i) := f;
            f := ai(j,i);
            ai(j,i) := ai(m,i);
            ai(m,i) := f;
         end loop;
      end exc;

   begin
      radix := Two;
      b2 := radix * radix;
      k := 1;
      l := n;

--     .......... search for rows isolating an eigenvalue
--                and push them down ..........
  <<L100>> null;

      for j in reverse 1 .. l loop
         allzero := True;
         for i in 1 .. l loop
            if (i /= j) then
               allzero := ar(j,i)=Zero and ai(j,i)=Zero;
               exit when not allzero;
            end if;
         end loop;
         if allzero then
            m := l;
            exc(m,j);
            if (l = 1) then
               low := k;
               igh := l;
               return;
            end if;
            l := l - 1;
            goto L100;
         end if;
      end loop;

--     .......... search for columns isolating an eigenvalue
--                and push them left ..........
  <<L140>> null;

      for j in k .. l loop
         allzero := True;
         for i in k .. L loop
            if (i /= j) then
               allzero := ar(i,j)=Zero and ai(i,j)=Zero;
               exit when not allzero;
            end if;
         end loop;
         if allzero then
            m := k;
            exc(m,j);
            k := k + 1;
            goto L140;
         end if;
      end loop;

--     .......... now balance the submatrix in rows k to l ..........
      for i in k .. l loop
         scale(i) := One;
      end loop;

--     .......... iterative loop for norm reduction ..........
  <<L190>> noconv := false;

      for i in k .. l loop
         c := Zero;
         r := Zero;
--
         for j in k .. l loop
            if (j /= i) then
               c := c + abs(ar(j,i)) + abs(ai(j,i));
               r := r + abs(ar(i,j)) + abs(ai(i,j));
            end if;
         end loop;

--     .......... guard against zero c or r due to underflow ..........
         if (c /= Zero) and (r /= Zero) then
            g := r / radix;
            f := One;
            s := c + r;
            while (c < g) loop
               f := f * radix;
               c := c * b2;
            end loop;
            g := r * radix;
            while (c >= g) loop
               f := f / radix;
               c := c / b2;
            end loop;

--     .......... now balance ..........
            if ((c + r) / f < 0.95 * s) then
               g := One / f;
               scale(i) := scale(i) * f;
               noconv := true;
               for j in k .. n loop
                  ar(i,j) := ar(i,j) * g;
                  ai(i,j) := ai(i,j) * g;
               end loop;
               for j in 1 .. l loop
                  ar(j,i) := ar(j,i) * f;
                  ai(j,i) := ai(j,i) * f;
               end loop;
            end if;
         end if;
      end loop;

      if noconv then goto L190; end if;
      low := k;
      igh := l;
   end cbal;

   procedure cbabk2(n,low,igh,m: in Integer; scale: in Vector; zr,zi: in out Matrix) is
     k: Integer;
     s: Rep;

   begin
      if (m = 0) then return; end if;

      if (igh /= low) then
         for i in low .. igh loop
            s := scale(i);
            for j in 1 .. m loop
               zr(i,j) := zr(i,j) * s;
               zi(i,j) := zi(i,j) * s;
            end loop;
         end loop;
      end if;

      for i in reverse 1 .. low-1 loop
         k := integer(scale(i));
         if (k /= i) then
            for j in 1 .. m loop
               s := zr(i,j);
               zr(i,j) := zr(k,j);
               zr(k,j) := s;
               s := zi(i,j);
               zi(i,j) := zi(k,j);
               zi(k,j) := s;
            end loop;
         end if;
      end loop;

      for i in igh+1 .. n loop
         k := integer(scale(i));
         if (k /= i) then
            for j in 1 .. m loop
               s := zr(i,j);
               zr(i,j) := zr(k,j);
               zr(k,j) := s;
               s := zi(i,j);
               zi(i,j) := zi(k,j);
               zi(k,j) := s;
            end loop;
         end if;
      end loop;
   end cbabk2;

   procedure comhes(n,low,igh: in Integer; ar,ai: in out Matrix; int: out Pivot) is
-- translated from eispack routine comhes, http://www.netlib.org/eispack/
      i,la,kp1,mm1,mp1: Integer;
      xr,xi,yr,yi: Rep;

   begin
      la := igh - 1;
      kp1 := low + 1;
--
      for m in kp1 .. la loop
         mm1 := m - 1;
         xr := Zero;
         xi := Zero;
         i := m;
--
         for j in m .. igh loop
            if (abs(ar(j,mm1)) + abs(ai(j,mm1)) > abs(xr) + abs(xi)) then
               xr := ar(j,mm1);
               xi := ai(j,mm1);
               i := j;
            end if;
         end loop;
--
         int(m) := i;
         if (i /= m) then

--     .......... interchange rows and columns of ar and ai ..........
            for j in mm1 .. n loop
               yr := ar(i,j);
               ar(i,j) := ar(m,j);
               ar(m,j) := yr;
               yi := ai(i,j);
               ai(i,j) := ai(m,j);
               ai(m,j) := yi;
            end loop;
--
            for j in 1 .. igh loop
               yr := ar(j,i);
               ar(j,i) := ar(j,m);
               ar(j,m) := yr;
               yi := ai(j,i);
               ai(j,i) := ai(j,m);
               ai(j,m) := yi;
            end loop;
         end if;

--     .......... end interchange ..........
         if (xr /= Zero) or (xi /= Zero) then
            mp1 := m + 1;
            for i in mp1 .. igh loop
               yr := ar(i,mm1);
               yi := ai(i,mm1);
               if (yr /= Zero) or (yi /= Zero) then
                  cdiv(yr,yi,xr,xi,yr,yi);
                  ar(i,mm1) := yr;
                  ai(i,mm1) := yi;
                  for j in m .. n loop
                     ar(i,j) := ar(i,j) - yr * ar(m,j) + yi * ai(m,j);
                     ai(i,j) := ai(i,j) - yr * ai(m,j) - yi * ar(m,j);
                  end loop;
                  for j in 1 .. igh loop
                     ar(j,m) := ar(j,m) + yr * ar(j,i) - yi * ai(j,i);
                     ai(j,m) := ai(j,m) + yr * ai(j,i) + yi * ar(j,i);
                  end loop;
               end if;
            end loop;
         end if;
      end loop;
   end comhes;

   procedure comlr2(n,low,igh: in Integer; int: in Pivot; hr,hi: in out Matrix;
                    wr,wi: out Vector; zr,zi: out Matrix; ierr: out Integer) is

-- translated from eispack routine comlr2, http://www.netlib.org/eispack/
      ii,l,m,en,im1,itn,its,mp1,enm1: Integer;
      si,sr,ti,tr,xi,xr,yi,yr,zzi,zzr,orm,tst1,tst2: Rep;

   begin
      ierr := 0;
--     .......... initialize eigenvector matrix ..........
      for i in  1 .. n loop
         for j in 1 .. n loop
            zr(i,j) := Zero;
            zi(i,j) := Zero;
         end loop;
         zr(i,i) := One;
      end loop;
--     .......... form the matrix of accumulated transformations
--                from the information left by comhes ..........
      for i in reverse low+1 .. igh-1 loop
         for k in i+1 .. igh loop
            zr(k,i) := hr(k,i-1);
            zi(k,i) := hi(k,i-1);
         end loop;
         ii := int(i);
         if (i /= ii) then
            for k in i .. igh loop
               zr(i,k) := zr(ii,k);
               zi(i,k) := zi(ii,k);
               zr(ii,k) := Zero;
               zi(ii,k) := Zero;
            end loop;
            zr(ii,i) := One;
         end if;
      end loop;

--     .......... store roots isolated by cbal ..........
      for i in 1 .. n loop
         if (i < low) or (i > igh) then
            wr(i) := hr(i,i);
            wi(i) := hi(i,i);
         end if;
      end loop;

      en := igh;
      tr := Zero;
      ti := Zero;
      itn := 30*n;

--     .......... search for next eigenvalue ..........
  <<L220>> if (en < low) then goto L680; end if;

      its := 0;
      enm1 := en - 1;

--     .......... look for single small sub-diagonal element
  <<L240>> null;

      for ll in reverse low .. en loop
         l := ll;
         exit when (l = low);
         tst1 := abs(hr(l-1,l-1)) + abs(hi(l-1,l-1)) + abs(hr(l,l)) + abs(hi(l,l));
         tst2 := tst1 + abs(hr(l,l-1)) + abs(hi(l,l-1));
         exit when (tst2 = tst1);
      end loop;

      if (l = en) then goto L660; end if;

      if (itn = 0) then
--     .......... set error -- all eigenvalues have not
--                converged after 30*n iterations ..........
        ierr := en;
        return;
      end if;

--     .......... form shift ..........
      if (its /= 10) and (its /= 20) then
         sr := hr(en,en);
         si := hi(en,en);
         xr := hr(enm1,en) * hr(en,enm1) - hi(enm1,en) * hi(en,enm1);
         xi := hr(enm1,en) * hi(en,enm1) + hi(enm1,en) * hr(en,enm1);
         if (xr /= Zero) or (xi /= Zero) then
            yr := (hr(enm1,enm1) - sr) / Two;
            yi := (hi(enm1,enm1) - si) / Two;
            csroot(yr**2-yi**2+xr,Two*yr*yi+xi,zzr,zzi);
            if (yr * zzr + yi * zzi < Zero) then
               zzr := -zzr;
               zzi := -zzi;
            end if;

            cdiv(xr,xi,yr+zzr,yi+zzi,xr,xi);
            sr := sr - xr;
            si := si - xi;
         end if;
      else
--     .......... form exceptional shift ..........
         sr := abs(hr(en,enm1)) + abs(hr(enm1,en-2));
         si := abs(hi(en,enm1)) + abs(hi(enm1,en-2));
      end if;

      for i in low .. en loop
         hr(i,i) := hr(i,i) - sr;
         hi(i,i) := hi(i,i) - si;
      end loop;

      tr := tr + sr;
      ti := ti + si;
      its := its + 1;
      itn := itn - 1;

--     .......... look for two consecutive small
--                sub-diagonal elements ..........
      xr := abs(hr(enm1,enm1)) + abs(hi(enm1,enm1));
      yr := abs(hr(en,enm1)) + abs(hi(en,enm1));
      zzr := abs(hr(en,en)) + abs(hi(en,en));

      for mm in reverse l .. enm1 loop
         m := mm;
         exit when (m = l);
         yi := yr;
         yr := abs(hr(m,m-1)) + abs(hi(m,m-1));
         xi := zzr;
         zzr := xr;
         xr := abs(hr(m-1,m-1)) + abs(hi(m-1,m-1));
         tst1 := zzr / yi * (zzr + xr + xi);
         tst2 := tst1 + yr;
         exit when (tst2 = tst1);
      end loop;

--     .......... triangular decomposition h:=l*r ..........
      mp1 := m + 1;

      for i in mp1 .. en loop
         im1 := i - 1;
         xr := hr(im1,im1);
         xi := hi(im1,im1);
         yr := hr(i,im1);
         yi := hi(i,im1);

         if (abs(xr) + abs(xi) < abs(yr) + abs(yi)) then
--     .......... interchange rows of hr and hi ..........
            for j in im1 .. n loop
               zzr := hr(im1,j);
               hr(im1,j) := hr(i,j);
               hr(i,j) := zzr;
               zzi := hi(im1,j);
               hi(im1,j) := hi(i,j);
               hi(i,j) := zzi;
            end loop;

            cdiv(xr,xi,yr,yi,zzr,zzi);
            wr(i) := One;
         else
            cdiv(yr,yi,xr,xi,zzr,zzi);
            wr(i) := -One;
         end if;

         hr(i,im1) := zzr;
         hi(i,im1) := zzi;

         for j in i .. n loop
            hr(i,j) := hr(i,j) - zzr * hr(im1,j) + zzi * hi(im1,j);
            hi(i,j) := hi(i,j) - zzr * hi(im1,j) - zzi * hr(im1,j);
         end loop;
      end loop;

--     .......... composition r*l:=h ..........
      for j in mp1 .. en loop
         xr := hr(j,j-1);
         xi := hi(j,j-1);
         hr(j,j-1) := Zero;
         hi(j,j-1) := Zero;

--     .......... interchange columns of hr, hi, zr, and zi, if necessary ..........
         if (wr(j) > Zero) then
            for i in 1 .. j loop
               zzr := hr(i,j-1);
               hr(i,j-1) := hr(i,j);
               hr(i,j) := zzr;
               zzi := hi(i,j-1);
               hi(i,j-1) := hi(i,j);
               hi(i,j) := zzi;
            end loop;

            for i in low .. igh loop
               zzr := zr(i,j-1);
               zr(i,j-1) := zr(i,j);
               zr(i,j) := zzr;
               zzi := zi(i,j-1);
               zi(i,j-1) := zi(i,j);
                zi(i,j) := zzi;
            end loop;
         end if;

         for i in 1 .. j loop
            hr(i,j-1) := hr(i,j-1) + xr * hr(i,j) - xi * hi(i,j);
            hi(i,j-1) := hi(i,j-1) + xr * hi(i,j) + xi * hr(i,j);
         end loop;

--     .......... accumulate transformations ..........
         for i in low .. igh loop
            zr(i,j-1) := zr(i,j-1) + xr * zr(i,j) - xi * zi(i,j);
            zi(i,j-1) := zi(i,j-1) + xr * zi(i,j) + xi * zr(i,j);
         end loop;
      end loop;

      goto L240;

--     .......... a root found ..........
  <<L660>> null;
      hr(en,en) := hr(en,en) + tr;
      wr(en) := hr(en,en);
      hi(en,en) := hi(en,en) + ti;
      wi(en) := hi(en,en);
      en := enm1;
      goto L220;

--     .......... all roots found.  backsubstitute to find
--                vectors of upper triangular form ..........
  <<L680>> null;
      orm := Zero;
      for i in 1 .. n loop
         for j in i .. n loop
            tr := abs(hr(i,j)) + abs(hi(i,j));
            if (tr > orm) then orm := tr; end if;
         end loop;
      end loop;

      hr(1,1) := orm;
      if (n = 1) or (orm = Zero) then return; end if;

      for nn in reverse 2 .. n loop
         en := nn;
         xr := wr(en);
         xi := wi(en);
         hr(en,en) := One;
         hi(en,en) := Zero;
         enm1 := en - 1;

         for i in reverse 1 .. enm1 loop
            zzr := Zero;
            zzi := Zero;

            for j in i+1 .. en loop
               zzr := zzr + hr(i,j) * hr(j,en) - hi(i,j) * hi(j,en);
               zzi := zzi + hr(i,j) * hi(j,en) + hi(i,j) * hr(j,en);
            end loop;

            yr := xr - wr(i);
            yi := xi - wi(i);
            if (yr = Zero) and (yi = Zero) then
               tst1 := orm;
               yr := Tst1;
               loop
                  yr := 1.0E-2 * yr;
                  tst2 := orm + yr;
                  exit when (tst2 <= tst1);
               end loop;
            end if;
            cdiv(zzr,zzi,yr,yi,hr(i,en),hi(i,en));

--     .......... overflow control ..........
            tr := abs(hr(i,en)) + abs(hi(i,en));
            if (tr /= Zero) then
               tst1 := tr;
               tst2 := tst1 + One/tst1;
               if (tst2 <= tst1) then
                  for j in i .. en loop
                     hr(j,en) := hr(j,en)/tr;
                     hi(j,en) := hi(j,en)/tr;
                  end loop;
               end if;
            end if;

         end loop;
      end loop;

--     .......... end backsubstitution ..........
--     .......... vectors of isolated roots ..........
      for i in 1 .. n loop
         if (i < low) or (i > igh) then
            for j in i .. n loop
               zr(i,j) := hr(i,j);
               zi(i,j) := hi(i,j);
            end loop;
         end if;
      end loop;

--     .......... multiply by transformation matrix to give
--                vectors of original full matrix.
      for j in reverse low .. n loop
         m := Integer'Min(j,igh);
--
         for i in low .. igh loop
            zzr := Zero;
            zzi := Zero;
            for k in low .. m loop
               zzr := zzr + zr(i,k) * hr(k,j) - zi(i,k) * hi(k,j);
               zzi := zzi + zr(i,k) * hi(k,j) + zi(i,k) * hr(k,j);
            end loop;
            zr(i,j) := zzr;
            zi(i,j) := zzi;
         end loop;
      end loop;
   end comlr2;

   procedure SortEigen(Eps: in Rep; Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
     Less1: constant Rep := One-Eps;
     X,Y: Integer;
     Ri,Rj: Rep;
     A: Pivot(Wr'Range);
     U: Vector(Wr'Range);

     function GT(I,J: Integer) return Boolean is
     begin
       Ri := sqr(Wr(I))+sqr(Wi(I));
       Rj := sqr(Wr(J))+sqr(Wi(J));
       if Less1*Ri > Rj then return  True; end if;
       if Less1*Rj > Ri then return False; end if;
       if Less1*Abs(Wr(I)) > Abs(Wr(J)) then return  True; end if;
       if Less1*Abs(Wr(J)) > Abs(Wr(I)) then return False; end if;
       if Wi(I) > Wi(J) then return  True; end if;
       return False;
     end GT;

     procedure Qsort(L,R: Integer) is
       I,J: integer;
     begin
       I:=L; J:=R; X:=A((L+R)/2);
       loop
         while GT(A(I),X) loop I:=I+1; end loop;
         while GT(X,A(J)) loop J:=J-1; end loop;
         if I <= J then
           Y:=A(I); A(I):=A(J); A(J):=Y;
           I:=I+1; J:=J-1;
         end if;
         exit when I>J;
       end loop;
       if L<J then Qsort(L,J); end if;
       if I<R then Qsort(I,R); end if;
     end Qsort;

   begin
     for J in Wr'Range loop A(J) := J; end loop;
     Qsort(Wr'First,Wr'Last);

     U := Wr;
     for J in Wr'Range loop Wr(J) := U(A(J)); end loop;
     U := Wi;
     for J in Wr'Range loop Wi(J) := U(A(J)); end loop;

     for I in Wr'Range loop
       for J in Wr'Range loop U(J) := Vr(I,J); end loop;
       for J in Wr'Range loop Vr(I,J) := U(A(J)); end loop;
       for J in Wr'Range loop U(J) := Vi(I,J); end loop;
       for J in Wr'Range loop Vi(I,J) := U(A(J)); end loop;
     end loop;
   end SortEigen;

   procedure FixEval(Wr,Wi: in out Vector; E: out Rep) is
--- assuming comlex conjugate eigenvalues are next to each other
     J: Integer := Wr'First;
     E1,E2: Rep;
   begin
     E := Zero;
     while J < Wr'Last loop
       E1 := abs(Wi(J));
       E2 := Rep'Max(abs(Wr(J)-Wr(J+1)),abs(Wi(J)+Wi(J+1)));
       if E1<E2 then
         Wi(J) := Zero;
         E := Rep'Max(E,E1/(abs(Wr(J))+E1));
         J := J+1;
       else
         Wr(J) := Half*(Wr(J)+Wr(J+1));
         Wr(J+1) := Wr(J);
         Wi(J) := Half*(Wi(J)-Wi(J+1));
         Wi(J+1) := -Wi(J);
         E := Rep'Max(E,E2/(abs(Wr(J))+E1));
         J := J+2;
       end if;
     end loop;
     if J=Wr'Last then
       E1 := abs(Wi(J));
       Wi(J) := Zero;
       E := Rep'Min(E,E1/(abs(Wr(J))+E1));
     end if;
   end FixEval;

   procedure FixEvec(Wi: in Vector; Vr,Vi: in out Matrix; E: out Rep) is
     J: Integer;
   begin
     E := Zero;
     for I in Wi'Range loop
       J := Wi'First;
       while J <= Wi'Last loop
         if Wi(J)=Zero then
           E := Rep'Max(E,abs(Vi(I,J)));
           Vi(I,J) := Zero;
           J := J+1;
         else
           E := Rep'Max(E,Rep'Max(abs(Vr(I,J)-Vr(I,J+1)),abs(Vi(I,J)+Vi(I,J+1))));
           Vr(I,J) := Half*(Vr(I,J)+Vr(I,J+1));
           Vr(I,J+1) := Vr(I,J);
           Vi(I,J) := Half*(Vi(I,J)-Vi(I,J+1));
           Vi(I,J+1) := -Vi(I,J);
           J := J+2;
         end if;
       end loop;
     end loop;
   end FixEvec;

   procedure NormalizeEvec(Wr,Wi: in Vector; Vr,Vi: in out Matrix) is
     K: Integer;
     S,E,C,Br,Bi: Rep;
   begin
     for J in Wr'Range loop
       S := Zero;
       E := Zero;
       for I in Wr'Range loop
         C := Sqr(Vr(I,J))+Sqr(Vi(I,J));
         S := S+C;
         if C>E then E := C; K := I; end if;
       end loop;

       Br := Vr(K,J);
       Bi := Vi(K,J);
       S := Sqrt(S*(Sqr(Br)+Sqr(Bi)));
       Br := Br/S;
       Bi := Bi/S;
       for I in Wr'Range loop
         S := Vr(I,J);
         C := Vi(I,J);
         Vr(I,J) := Br*S+Bi*C;
         Vi(I,J) := Br*C-Bi*S;
       end loop;
       Vi(K,J) := Zero;
     end loop;
   end NormalizeEvec;

  procedure Cleanup(W: in out Vector) is
    Small: constant Rep := 1.0E-200;
  begin
    for I in W'Range loop
      if abs(W(I))<Small then W(I) := Zero; end if;
    end loop;
  end Cleanup;

  procedure Cleanup(V: in out Matrix) is
    Small: constant Rep := 1.0E-100;
  begin
    for I in V'Range(1) loop
      for J in V'Range(2) loop
        if abs(V(I,J))<Small then V(I,J) := Zero; end if;
      end loop;
    end loop;
  end Cleanup;

  procedure Eigen(A: in Matrix; Wr,Wi: out Vector; Vr,Vi: out Matrix) is
    pragma Suppress(Storage_Check);
    Eps: constant Rep := 1.0E-8;
    Err: constant Rep := 1.0E-10;
    N: constant Integer := A'Last(1);
    Low,Igh,Ierr: Integer;
    E: Rep;
    Int: Pivot(1 .. N);
    Scale: Vector(1 .. N);
    Ar,Ai: Matrix(1 .. N,1 .. N);
  begin
    TraceEnter("Linear.Ops.Eigen");
    Ar := A;
    SetZero(Ai);
    Cbal(N,Ar,Ai,Low,Igh,Scale);
    Comhes(N,Low,Igh,Ar,Ai,Int);
    Comlr2(N,Low,Igh,Int,Ar,Ai,Wr,Wi,Vr,Vi,Ierr);
    if Ierr>0 then
      Show("Error: Problem with eigenvalue number ",Ierr);
    end if;
    Cbabk2(N,Low,Igh,N,Scale,Vr,Vi);
    SortEigen(Eps,Wr,Wi,Vr,Vi);
    FixEval(Wr,Wi,E);
    if E>Err then
      Show("Error in FixEval: ",E);
    end if;
    NormalizeEvec(Wr,Wi,Vr,Vi);
    FixEvec(Wi,Vr,Vi,E);
    if E>Err then
      Show("Error in FixEvec: ",E);
    end if;
    Cleanup(Wr); Cleanup(Wi);
    Cleanup(Vr); Cleanup(Vi);
    TraceLeave;
  end Eigen;

end Linear.Ops;
