package Linear.Ops is

  pragma Elaborate_Body;

  function CountLarge(B: Vector; Cut: Rep := Zero) return Integer;       -- Number of elements larger than Cut

  procedure Add(B1: in Vector; B2: in out Vector);                       -- B2 := B2+B1
  procedure Add(A1: in Matrix; A2: in out Matrix);                       -- A2 := A2+A1
  procedure Add(R: in Rep; A: in out Matrix);                            -- A := A+R*Identity
  procedure Sub(B1: in Vector; B2: in out Vector);                       -- B2 := B2-B1
  procedure Sub(A1: in Matrix; A2: in out Matrix);                       -- A2 := A2-A1
  procedure Mult(R: in Rep; B: in out Vector);                           -- B := R*B
  procedure Mult(R: in Rep; A: in out Matrix);                           -- A := R*A
  procedure Mult(A1: in Matrix; A2: in out Matrix);                      -- A2 := A1*A2
  function Transpose(A: Matrix) return Matrix;

  function "-"(B: Vector) return Vector;                                 -- -B
  function "+"(B1,B2: Vector) return Vector;                             -- B1+B2
  function "+"(A1,A2: in Matrix) return Matrix;                          -- A1+A2
  function "-"(B1,B2: Vector) return Vector;                             -- B1-B2
  function "-"(A1,A2: in Matrix) return Matrix;                          -- A1-A2
  function "*"(B1,B2: Vector) return Rep;                                -- L2 inner product <B1,B2>
  function "*"(R: Rep; B: Vector) return Vector;                         -- R*B
  function "*"(R: Rep; A: Matrix) return Matrix;                         -- R*A
  function "*"(B1,B2: Vector) return Vector;                             -- Diag(B1)*B2
  function "*"(A: Matrix; B: Vector) return Vector;                      -- A*B
  function "*"(B: Vector; A: Matrix) return Vector;                      -- Transpose(A)*B
  function "*"(A1,A2: Matrix) return Matrix;                             -- A1*A2

  procedure Solve(A: in Matrix; B: in Vector; X: out Vector);            -- solve A*X=B
  procedure TSolve(A: in Matrix; B: in Vector; X: out Vector);           -- solve transpose(A)*X=B
  function Det(A: Matrix) return Rep;                                    -- determinant of A
  procedure Invert(A: in out Matrix);                                    -- invert A
  procedure Eigen(A: in Matrix; W: out Vector);                          -- eigenvalues of symmetric A
  procedure Eigen(A: in Matrix; W: out Vector; V: out Matrix);           -- eigenvalues/eigenvectors of symmetric A
  procedure Eigen(A: in Matrix; Wr,Wi: out Vector; Vr,Vi: out Matrix);   -- eigenvalues/eigenvectors of general A
  procedure Cleanup(W: in out Vector);
  procedure Cleanup(V: in out Matrix);

private

-- procedures translated/adapted from linpack

  procedure dgefa(n: in Integer; a: in out Matrix; ipvt: out Pivot; info: out Integer);
  procedure dgesl(n,job: in Integer; ipvt: in Pivot; a: in Matrix; b: in out Vector);
  procedure dgedi(n,job: in Integer; ipvt: in Pivot; a: in out Matrix; work: in out Vector; det1,det2: out Rep);

-- procedures translated/adapted from eispack

  function dsign(a,b: Rep) return Rep;

  procedure tred1(n: in Integer; a: in out Matrix; d,e,e2: out Vector);
  procedure tred2(n: in Integer; a: in Matrix; d,e: out Vector; z: out Matrix);
  procedure tql1(n: in Integer; d,e: in out Vector; ierr: out Integer);
  procedure tql2(n: in Integer; d,e: in out Vector; z: in out Matrix; ierr: out Integer);

  procedure SortEigen(W: in out Vector; V: in out Matrix);

  procedure cdiv(ar,ai,br,bi: in Rep; cr,ci: out Rep);
  procedure csroot(xr,xi: in Rep; yr,yi: out Rep);
  procedure cbal(n: in Integer; ar,ai: in out Matrix; low,igh: out Integer; scale: out Vector);
  procedure cbabk2(n,low,igh,m: in Integer; scale: in Vector; zr,zi: in out Matrix);
  procedure comhes(n,low,igh: in Integer; ar,ai: in out Matrix; int: out Pivot);
  procedure comlr2(n,low,igh: in Integer; int: in Pivot; hr,hi: in out Matrix;
                   wr,wi: out Vector; zr,zi: out Matrix; ierr: out Integer);

  procedure SortEigen(Eps: in Rep; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);
  procedure FixEval(Wr,Wi: in out Vector; E: out Rep);
  procedure FixEvec(Wi: in Vector; Vr,Vi: in out Matrix; E: out Rep);
  procedure NormalizeEvec(Wr,Wi: in Vector; Vr,Vi: in out Matrix);

end Linear.Ops;
