#include <fenv.h>

#define STRICT_MODE   1
#define RELAXED_MODE  2
#define FE_STRICT (FE_DIVBYZERO | FE_UNDERFLOW | FE_OVERFLOW | FE_INVALID)
#define u(x) if (x != 0) abort()
#define v(x) if (x == -1) abort()

void modify_fp_env(int new_mode) {
  static int saved_mode = 0;
  static fexcept_t saved_fexcept;

  if (saved_mode == 0) {                                /* initialize */
    u(feclearexcept(FE_ALL_EXCEPT));
    u(fegetexceptflag(&saved_fexcept,FE_ALL_EXCEPT));   /* save exception flags */
  }
  if (new_mode != saved_mode) {                         /* switch mode */
    if (new_mode == STRICT_MODE) {                      /* strict mode */
      u(fesetexceptflag(&saved_fexcept,FE_ALL_EXCEPT)); /* restore exception flags */
      v(feenableexcept(FE_STRICT));                     /* enable strict checking */
      u(fesetround(FE_UPWARD));                         /* enter round-up mode */
      saved_mode = STRICT_MODE;
    }
    else {                                              /* relaxed mode */
      u(fegetexceptflag(&saved_fexcept,FE_ALL_EXCEPT)); /* save exception flags */
      v(fedisableexcept(FE_ALL_EXCEPT));                /* mask all exceptions */
      u(fesetround(FE_TONEAREST));                      /* enter round-to-nearest mode */
      saved_mode = RELAXED_MODE;
    }
  }
}

