with Ada.Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Move is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;

   Size:     constant Positive := GetArg(1); -- Maximal frequency for explicit Fourier polynomials
   PartSize: constant Positive := GetArg(2); -- Cutoff for matrix approximations
   QName:    constant String   := GetArg(3); -- filename of approximate solution
   DA:       constant Rep      := GetArg(4); -- delta alpha
   DL:       constant Rep      := GetArg(5); -- delta lambda
   R1:       constant Radius   := 0.1;       -- radius of analyticity

   Number: Positive;
   F: File_Type;
   Alpha : Scalar;

begin
   Open(F,In_File,QName);
   Get(F,Alpha,False);
   Txt_Int_IO.Get(F,Number);
   Close(F);

   declare
      package NKs is new Ks (Number => Number, Size => Size, PartSize => PartSize, Scalar => Scalar);
      use NKs, NKs.Qf, NKs.Qf_Ops, NKs.Qf_IO;

      Eps: constant Radius := 1.0E-18;
      G,H: Q;
      Alpha0,Lambda0: Scalar;

   begin
      Read(QName,1,Alpha0,H,False);
      G := ApproxBifVec(Alpha0,H);
      Lambda0 := P1(H,G);
      Alpha0 := Alpha0+Scal(DA);
      NewtonU(Alpha0,Lambda0+Scal(DL),G,Eps,H);
--      Newton(Alpha0,Eps,H);
      Write(QName & "-final",Alpha0,H,False);
      GuessDomain(H);
   end;

end Move;

