with Ada.Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
use Ada.Text_IO, Messages, Reps, Linear, Linear.IO, Linear.Ops;
with Numerics, Numerics.Ops, Numerics.IO;
use  Numerics, Numerics.Ops, Numerics.IO;
with Ks;

procedure Newton is

-- this is an implementation of the Newton algorithm
-- it tries to locate a solution with tolerance Eps
-- close to a given function

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   package Txt_Rep_IO is new Ada.Text_IO.Float_IO(Rep);

   subtype Scalar is Numeric;

   Size:     constant Positive := GetArg(1); -- Maximal frequency for explicit Fourier polynomials
   PartSize: constant Positive := GetArg(2); -- Cutoff for matrix approximations
   QName:    constant String   := GetArg(3); -- filename of approximate solution
   DecIO:    constant Boolean  := GetArg(4); -- decimal input/output for Scalar
   Eps :     constant Radius   := GetArg(5); -- tolerance
   R1  :  constant Radius  := 0.1;           -- radius of analyticity

   Number: Positive;
   F: File_Type;
   Alpha: Scalar;

begin
   Open(F,In_File,QName);
   Get(F,Alpha,DecIO);
   Txt_Int_IO.Get(F,Number);
   Close(F);

   declare
      package NKs is new Ks (Number => Number, Size => Size, PartSize => PartSize, Scalar => Scalar);
      use NKs, NKs.Qf, NKs.Qf_Ops, NKs.Qf_IO;

      H: Q;

   begin
      ShowParam;
      Read(QName,1,Alpha,H,DecIO);
      Show("alpha:",Approx(Alpha));
      Newton(Alpha,Eps,H);
      Write(QName & "-final",Alpha,H,DecIO);
      GuessDomain(H);
   end;

end Newton;
