with Reps;
use Reps;

package Numerics is

  pragma Pure;

  type Numeric is new Rep;

  function IsNumeric(S: Numeric) return Boolean;
  function Scal(K: Integer) return Numeric;
  function Scal(R: Rep) return Numeric;
  function Scal(R1,R2: Rep) return Numeric;
  function Ball(R: Rep) return Numeric;
  function Ball(S: Numeric) return Numeric;
  function Approx(S: Numeric) return Rep;
  procedure Split(S: in Numeric; S0,SE: out Numeric);
  function Piece(M,N: Positive; S: Numeric) return Numeric;
  function Inf(S: Numeric) return Rep renames Approx;
  function Sup(S: Numeric) return Rep renames Approx;
  function InfAbs(S: Numeric) return Rep;
  function SupAbs(S: Numeric) return Rep renames InfAbs;
  function Min(S1,S2: Numeric) return Numeric renames Numeric'Min;
  function Max(S1,S2: Numeric) return Numeric renames Numeric'Max;
  function Envelope(S1,S2: Numeric) return Numeric;
  function Sqr(S: Numeric) return Numeric;
  function "*"(R: Rep; S: Numeric) return Numeric;
  function Inv(S: Numeric) return Numeric;
  function "/"(S: Numeric; R: Rep) return Numeric;

  pragma Inline (Scal,Ball,Approx,Split,Inf,Sup,SupAbs,Max,Sqr,"*","/");
  pragma Inline_Always (Scal,Ball,Approx,Split,Inf,Sup,SupAbs,Max,Sqr,"*","/");

end Numerics;
