with Messages, Numerics, Numerics.Ops, Numerics.IO, Reps.Ops;
use Messages, Numerics, Numerics.Ops, Numerics.IO;
with Basics;

package body Packed is

   package Basic is new Basics (Number => Number, Size => Size, PartSize => PartSize, Scalar => Numeric);
   use Basic, Basic.Qb, Basic.Qb_Ops, Basic.Qb_IO;

   function DF(Alpha: Rep; J: Parity; R: Radius; A: Vector) return Matrix is
     H: Q;
   begin
     UnPack(A,J,R,H);
     return DF(Scal(Alpha),H);
   end DF;

   function DFmI(Alpha: Rep; J: Parity; R: Radius; A: Vector) return Matrix is
     H: Q;
   begin
     UnPack(A,J,R,H);
     return DFmI(Scal(Alpha),H);
   end DFmI;

   function DUtildemI(Alpha: Rep; J: Parity; R: Radius; A,B: Vector) return Matrix is
     H,G: Q;
   begin
     UnPack(A,J,R,H);
     UnPack(B,J,R,G);
     return DUtildemI(Scal(Alpha),H,G);
   end DUtildemI;

   function L(Alpha: Rep; J: Parity; R: Radius; A: Vector) return Matrix is
     H: Q;
   begin
     UnPack(A,J,R,H);
     return L(Scal(Alpha),H);
   end L;

   function Wi(Alpha: Rep; J: Parity; R: Radius; B0,A: Vector) return Matrix is
    H0,G: Q;
   begin
     Unpack(B0,J,R,H0);
     Unpack(A,J,R,G);
     return Wi(Scal(Alpha),H0,G);
   end Wi;

   procedure Newton(Alpha: in Rep; Eps: in Radius; J: in Parity; R: in Radius; A: in out Vector) is
     pragma Suppress(Storage_Check);
     use Reps.Ops;
     Nmax: constant Integer := 32;
     Ratio: constant Rep := Half;
     E,Et: Rep := Eps;
     G,H: Q;
     M: Matrix(PartIndex,PartIndex);
   begin
     UnPack(A,J,R,H);
     for N in 1 .. Nmax loop
        CleanUp(H);
        G := F(Scal(Alpha),H);
        Et := E;
        E := Sup(Norm(G-H));
        if E<Eps then                          -- done
           Pack(H,A);
           return;
        end if;
        if E/Et>Ratio or E>Sqrt(Eps) then      -- need to update M
           M := DFmI(Scal(Alpha),H);
        end if;
        H := G-M*(G-H);
     end loop;
--     Show("Fpu_Packed.Newton: giving up");
     Pack(H,A);
   end Newton;

   procedure NewtonU(Alpha,Lambda: in Rep; W: in Vector; Eps: in Radius; J: in Parity; R: in Radius; A: in out Vector) is
     pragma Suppress(Storage_Check);
     use Reps.Ops;
     Nmax: constant Integer := 32;
     Ratio: constant Rep := Half;
     E,Et: Rep := Eps;
     G,H,Ut: Q;
     M: Matrix(PartIndex,PartIndex);
   begin
     UnPack(W,J,R,Ut);
     UnPack(A,J,R,H);
     H := P2(H,Ut)+Scal(Lambda)*Ut;
     for N in 1 .. Nmax loop
        G :=  Utilde(Scal(Alpha),Scal(Lambda),H,Ut);
        Et := E;
        E := Sup(Norm(G-H));
        if E<Eps then
           Pack(H,A);
           return;
        end if;
        if E/Et>Ratio or E>Sqrt(Eps) then
           M := DUtildemI(Scal(Alpha),H,Ut);
        end if;
        H := G-M*(G-H);
     end loop;
     Show("Fpu_Packed.NewtonU: giving up");
     Pack(H,A);
   end NewtonU;

end Packed;
