with Messages;
use Messages;

pragma Elaborate_All (Messages);

package body Qs.Ops is

  ScalZero: constant Scalar  := Scal(Zero);
  Trunc:    constant Boolean := IsNumeric(ScalZero);

  function Comp(N: Freq; H: Fourier) return Scalar renames Components.Component;
  function ErrComp(N: Freq; H: Fourier) return Scalar renames Components.ErrComponent;
  procedure SetComp(N: Freq; S: in Scalar; H: in out Fourier) renames Components.SetComponent;
  procedure SetErrComp(N: FreqE; S: in Scalar; H: in out Fourier) renames Components.SetErrComponent;
  function Pak(N: Freq; M: Part) return Positive renames Pack;
  procedure UnPak(K: in Positive; N: out Freq; M: out Part) renames UnPack;

  pragma Inline (Pak,UnPak,Comp,ErrComp,SetComp,SetErrComp);

  function "+"(H1,H2: Q) return Q is
     H: Q;
  begin
     for M in Part loop
        H(M) := H1(M)+H2(M);
     end loop;
     return H;
  end "+";

  function "-"(H1,H2: Q) return Q is
     H: Q;
  begin
     for M in Part loop
        H(M) := H1(M)-H2(M);
     end loop;
     return H;
  end "-";

  function "-"(H: Q) return Q is
     Ht: Q;
  begin
     for M in Part loop
        Ht(M) := -H(M);
     end loop;
     return Ht;
  end "-";

  function "*"(H1,H2: Q) return Q is
    H: Q;
  begin
     for M in Part loop
        H(M) := H1(M)*H2(M);
     end loop;
     return H;
  end "*";

  function "*"(H1,H2: Q) return Scalar is
    S: Scalar := ScalZero;
  begin
     for M in Part loop
        S := S+H1(M)*H2(M);
     end loop;
     return S;
  end "*";

  function "*"(S: Scalar; H: Q) return Q is
     Ht: Q;
  begin
     for M in Part loop
        Ht(M) := S*H(M);
     end loop;
     return Ht;
  end "*";

  function "*"(A: Matrix; H: Q) return Q is
---  all errors of the result are in the coefficients
     N1,N2,Nmax: Freq;
     M1,M2: Part;
     K: Integer;
     C,S: Scalar;
     Wi: array(Freq) of Scalar;
     G: Q;
  begin
     if not Poly(H) then
        Wi(0) := Scal(One);
        Wi(1) := Exp(Scal(-Rho(H)));
        K := 1;
        for N in 2 .. Size loop
           Wi(N) := Wi(N-K)*Wi(K);                -- Exp(-N*Rho(H))
           if 2*K=N then K := N; end if;
        end loop;
     end if;

     SetZero(Par(H),Rho(H),G);
     UnPak(A'Last(2),Nmax,M2);
     for K1 in A'Range(1) loop
        UnPak(K1,N1,M1);
        C := ScalZero;
        for K2 in A'Range(2) loop                 -- coeff -> coeff
           UnPak(K2,N2,M2);
           C := C+A(K1,K2)*Comp(N2,H(M2));
        end loop;
        if not Poly(H) then                       -- errors -> coeff
           for M in Part loop
              S := ScalZero;
              for N in reverse Par(H) .. Nmax loop
                 K := Pak(N,M);                   -- should be in A'Range(2)
                 S := Max(S,Abs(A(K1,K))*Wi(N));
                 C := C+S*Ball(ErrComp(N,H(M)));
              end loop;
           end loop;
        end if;
        SetComp(N1,C,G(M1));
     end loop;

     return G;
  end "*";

  function "*"(A: SMatrix; H: Q) return Q is
---  all errors of the result are in the coefficients
     N1,N2,Nmax: Freq;
     M1,M2: Part;
     K: Integer;
     C,S: Scalar;
     Wi: array(Freq) of Scalar;
     G: Q;
  begin
     if not Poly(H) then
        Wi(0) := Scal(One);
        Wi(1) := Exp(Scal(-Rho(H)));
        K := 1;
        for N in 2 .. Size loop
           Wi(N) := Wi(N-K)*Wi(K);               -- powers of 1/Rho(H)
           if 2*K=N then K := N; end if;
        end loop;
     end if;

     SetZero(Par(H),Rho(H),G);
     UnPak(A'Last(2),Nmax,M2);

     for K1 in A'Range(1) loop
        UnPak(K1,N1,M1);
        C := ScalZero;
        for K2 in A'Range(2) loop                 -- coeff -> coeff
           UnPak(K2,N2,M2);
           C := C+A(K1,K2)*Comp(N2,H(M2));
        end loop;
        if not Poly(H) then                       -- errors -> coeff
           for M in Part loop
              S := ScalZero;
              for N in reverse Par(H) .. Nmax loop
                 K := Pak(N,M);                   -- should be in A'Range(2)
                 S := Max(S,Abs(A(K1,K))*Wi(N));
                 C := C+S*Ball(ErrComp(N,H(M)));
              end loop;
           end loop;
        end if;
        SetComp(N1,C,G(M1));
     end loop;
     return G;
  end "*";

   function Ball(H: Q; Eps: Radius) return Q is
     E: Scalar;
     G: Q := H;
   begin
     for M in Part loop
       E := ErrComp(Par(G(M)),G(M))+Scal(Eps);
       SetErrComp(Par(G(M)),E,G(M));
     end loop;
     return G;
   end Ball;

  function Norm(G: Q) return Scalar is
     S: Radius := Zero;
     T: Radius := 1.0E+100;
  begin
     for M in Part loop
        S := Rep'Max(S,Sup(Norm(G(M))));
        T := Rep'Min(T,Inf(Norm(G(M))));
     end loop;
     return Scal(T,S);
  end Norm;

  function L2Norm(G: Q) return Scalar is
     S: Scalar := ScalZero;
  begin
     for M in Part loop
        S := S+Sqr(L2Norm(G(M)));
     end loop;
     return Sqrt(S);
  end L2Norm;

  function D(K: Positive; G: Q) return Q is
    H: Q;
  begin
    for M in Part loop
      H(M) := D(K,G(M));
    end loop;
    return H;
  end D;

  function D(G: Q; R: Radius) return Q is
    H: Q;
  begin
    for M in Part loop
      H(M) := D(G(M),R);
    end loop;
    return H;
  end D;

  function Dinv(B,C: Scalar; G: Q) return Q is
    H: Q;
  begin
    for M in Part loop
      H(M) := Dinv(B,C,G(M));
    end loop;
    return H;
  end Dinv;

  function Dinv(K: Positive; C: Scalar; G: Q) return Q is
    H: Q;
  begin
    for M in Part loop
      H(M) := Dinv(K,C,G(M));
    end loop;
    return H;
  end Dinv;

  function Dinv(K: Positive; B,C: Scalar; G: Q) return Q is
    H: Q;
  begin
    for M in Part loop
      H(M) := Dinv(K,B,C,G(M));
    end loop;
    return H;
  end Dinv;

  procedure Split(N: in Freq; H: in Q; Hl,Hh: out Q) is
  begin
    for M in Part loop
      Split(N,H(M),Hl(M),Hh(M));
    end loop;
  end Split;

  procedure SplitLowErr(N: in Freq; H: in out Q; E: out Scalar) is
    S: Scalar;
  begin
    E := ScalZero;
    for M in Part loop
      SplitLowErr(N,H(M),S);
      E := Max(S,E);
    end loop;
  end SplitLowErr;

  procedure Split(H: in Q; Hl,Hh: out Q) is
  begin
    for M in Part loop
      Split(H(M),Hl(M),Hh(M));
    end loop;
  end Split;

  procedure ZeroConstant(H: in out Q) is
  begin
    for M in Part loop
      ZeroConstant(H(M));
    end loop;
  end ZeroConstant;

  function ZeroConstant(H: Q) return Q is
    G: Q := H;
  begin
    ZeroConstant(G);
    return G;
  end ZeroConstant;

  function P1(H,G: Q) return Scalar is
  begin
    return (H*G)/Sqrt(G*G);
  end P1;

  function P1(H,G: Q) return Q is
  begin
    return ((H*G)/(G*G))*G;
  end P1;

  function P2(H,G: Q) return Q is
  begin
     return H-P1(H,G);
  end P2;

  procedure GuessDomain(H: in Q) is
    R: Rep := Rep'Last;
    Rk: Rep;
  begin
     for M in Part loop
        GuessRho(H(M),Rk);
        R := Rep'Min(R,Rk);
     end loop;
     Show("H seems analytic for |Im z| <",R);
     Show("H is used for |Im z| <",Rho(H));
  end GuessDomain;

  procedure CleanUp(H: in out Q) is
  begin
     for M in Part loop
        CleanUp(H(M));
     end loop;
  end CleanUp;

  function Alpha4(H: Q) return Q is
     G: Q;
  begin
     for M in Part loop
        G(M) := Alpha4(H(M));
     end loop;
     return G;
  end Alpha4;

end Qs.Ops;
