with Fouriers.Ops, SLinear.Ops;

pragma Elaborate_All (Fouriers.Ops);

generic

  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with function "*"(R: Rep; S: Scalar) return Scalar is <>;
  with function "/"(S: Scalar; R: Rep) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;

package Qs.Ops is

  package Components_Ops is new Components.Ops;
  package QSLinear_Ops is new QSLinear.Ops;
  use Components_Ops, QSLinear, QSLinear_Ops;

  function "+"(H1,H2: Q) return Q;
  function "-"(H1,H2: Q) return Q;
  function "-"(H: Q) return Q;
  function "*"(H1,H2: Q) return Q;                        -- mult componentwise
  function "*"(H1,H2: Q) return Scalar;                   -- inner product
  function "*"(S: Scalar; H: Q) return Q;
  function "*"(A: Matrix; H: Q) return Q;
  function "*"(A: SMatrix; H: Q) return Q;
  function Ball(H: Q; Eps: Radius) return Q;              -- Eps-Ball centered at H
  function Norm(G: Q) return Scalar;
  function L2Norm(G: Q) return Scalar;
  function D(K: Positive; G: Q) return Q;                 -- D^N G  (polynomials)
  function D(G: Q; R: Radius) return Q;                   -- D G
  function Dinv(B,C: Scalar; G: Q) return Q;              -- C*[1+B*D^(-2)]^(-1) G
  function Dinv(K: Positive; C: Scalar; G: Q) return Q;   -- C*D^(-K) G
  function Dinv(K: Positive; B,C: Scalar; G: Q) return Q; -- C*[D^K(D^2+B)]^(-1) G
  procedure ZeroConstant(H: in out Q);                    -- sets constant part =0
  function ZeroConstant(H: Q) return Q;                   -- sets constant part =0
  procedure Split(N: in Freq; H: in Q; Hl,Hh: out Q); -- split into low/high modes
  procedure SplitLowErr(N: in Freq; H: in out Q; E: out Scalar);
                                           -- split off error terms for low modes
  procedure Split(H: in Q; Hl,Hh: out Q);  -- split into poly/errors
  function P1(H,G: Q) return Scalar;       -- projects H in the direction of G
  function P1(H,G: Q) return Q;            -- projects H in the direction of G
  function P2(H,G: Q) return Q;            -- projects H in the dir orth to G
  procedure GuessDomain(H: in Q);
  procedure CleanUp(H: in out Q);
  function Alpha4(H: Q) return Q;

end Qs.Ops;
