with Messages;
use Messages;

package body Qs is

   function Comp(N: Freq; H: Fourier) return Scalar renames Components.Component;
   function ErrComp(N: Freq; H: Fourier) return Scalar renames Components.ErrComponent;
   procedure SetComp(N: Freq; S: in Scalar; H: in out Fourier) renames Components.SetComponent;
   procedure SetErrComp(N: FreqE; S: in Scalar; H: in out Fourier) renames Components.SetErrComponent;
   function Pak(N: Freq; M: Part) return Positive renames Pack;
   procedure UnPak(K: in Positive; N: out Freq; M: out Part) renames UnPack;

   pragma Inline (Pak,UnPak,Comp,ErrComp,SetComp,SetErrComp);

   procedure SetZero(J: in Parity; R: in Radius; H: out Q) is
   begin
      SetZero(J,R,H(1));
      for M in 2 .. Number loop
         H(M) := H(1);
      end loop;
   end SetZero;

   function Par(H: Q) return Parity is
   begin
      return Par(H(1));
   end Par;

   function Rho(H: Q) return Radius is
   begin
      return Rho(H(1));
   end Rho;

   procedure SetRho(R: in Radius; H: in out Q) is
   begin
      for M in Part loop
         SetRho(R,H(M));
      end loop;
   end SetRho;

   function Poly(H: Q; Check: in Boolean := False) return Boolean is
   begin
     if Trunc then return True; end if;
     for M in Part loop
       if not Poly(H(M),Check) then return False; end if;
     end loop;
     return True;
   end Poly;

   function Component(M: Part; H: Q) return Fourier is
   begin
      return H(M);
   end Component;

   function SubComponent(N: Freq; M: Part; H: Q) return Scalar is
   begin
      return Comp(N,H(M));
   end SubComponent;

   function ErrComponent(N: FreqE; M: Part; H: Q) return Scalar is
   begin
      return ErrComp(N,H(M));
   end ErrComponent;

   procedure SetComponent(M: in Part; F: in Fourier; H: in out Q) is
   begin
      H(M) := F;
   end SetComponent;

   procedure SetSubComponent(N: in Freq; M: in Part; S: in Scalar; H: in out Q) is
   begin
      SetComp(N,S,H(M));
   end SetSubComponent;

   procedure SetErrComponent(N: FreqE; M: Part; S: in Scalar; H: in out Q) is
   begin
      SetErrComp(N,S,H(M));
   end SetErrComponent;

   function Pack(N: Freq; M: Part) return Positive is
   begin
      return M+(N-1)*Number;
   end Pack;

   procedure UnPack(K: in Positive; N: out Freq; M: out Part) is
   begin
      M := 1+ (K-1) mod Number;
      N := 1+ (K-1)/Number;
   end UnPack;

   procedure Pack(H: in Q; A: out Vector) is
      N: Freq;
      M: Part;
   begin
      for K in A'Range loop
         UnPak(K,N,M);
         A(K) := Approx(Comp(N,H(M)));
      end loop;
   end Pack;

   procedure UnPack(A: in Vector; J: in Parity; R: in Radius; H: out Q) is
      N: Freq;
      M: Part;
   begin
      if J=0 then
         Message("Qs.Ops.UnPack",Not_Implemented);
      end if;
      SetZero(J,R,H);
      for K in A'Range loop
         UnPak(K,N,M);
         SetComp(N,Scal(A(K)),H(M));
      end loop;
   end UnPack;

end Qs;
