with Reps,Fouriers_Init, Fouriers, Linear, SLinear;
use Reps, Fouriers_Init, Linear;

generic

  Number,Size: in Positive;

  type Scalar is private;

  with function IsNumeric(S: Scalar) return Boolean is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function Scal(R: Rep) return Scalar is <>;
  with function Scal(R1,R2: Rep) return Scalar is <>;
  with function Ball(R: Rep) return Scalar is <>;
  with function Ball(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function Approx(S: Scalar) return Rep is <>;
  with function Inf(S: Scalar) return Rep is <>;
  with function Sup(S: Scalar) return Rep is <>;
  with function InfAbs(S: Scalar) return Rep is <>;
  with function SupAbs(S: Scalar) return Rep is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Envelope(S1,S2: Scalar) return Scalar is <>;

package Qs is

  package Components is new Fouriers (Size => Size, Scalar => Scalar);
  use Components;

  package QSLinear is new SLinear(Scalar => Scalar);

  subtype Part is Integer range 1 .. Number;
  type Q is private;

  procedure SetZero(J: in Parity; R: in Radius; H: out Q);         -- H := 0
  function Par(H: Q) return Parity;                                -- Parity of H  (0=even 1=odd)
  function Rho(H: Q) return Radius;
  procedure SetRho(R: in Radius; H: in out Q);                     -- set radius on components
  function Poly(H: Q; Check: in Boolean := False) return Boolean;  -- H is a Fourier polynomial
  function Component(M: Part; H: Q) return Fourier;                -- H[M]
  function SubComponent(N: Freq; M: Part; H: Q) return Scalar;     -- Nth component of H[M]
  function ErrComponent(N: FreqE; M: Part; H: Q) return Scalar;    -- Nth error of H[M]
  procedure SetComponent(M: in Part; F: in Fourier; H: in out Q) ; -- H[M] := S
  procedure SetSubComponent(N: in Freq; M: in Part; S: in Scalar; H: in out Q);
  procedure SetErrComponent(N: FreqE; M: Part; S: in Scalar; H: in out Q); -- sets error component
  procedure Pack(H: in Q; A: out Vector);
  procedure UnPack(A: in Vector; J: in Parity; R: in Radius; H: out Q);
  function Pack(N: Freq; M: Part) return Positive;
  procedure UnPack(K: in Positive; N: out Freq; M: out Part);

private

  type Q is array(Part) of Fourier;

  ScalZero: constant Scalar := Scal(Zero);
  Trunc: constant Boolean := IsNumeric(ScalZero);

end Qs;
