with Ada.Numerics.Generic_Elementary_Functions;

pragma Elaborate_All(Ada.Numerics.Generic_Elementary_Functions);

package body Reps.Ops is

  function Sqr(R: Rep) return Rep is
  begin
    return R*R;
  end Sqr;

  package Rep_EF is new Ada.Numerics.Generic_Elementary_Functions(Rep);

  function Sqrt(R: Rep) return Rep renames Rep_EF.Sqrt;
  function Log(R: Rep) return Rep renames Rep_EF.Log;
  function Exp(R: Rep) return Rep renames Rep_EF.Exp;

  procedure Root4(C0,C1,C3: in Rep; R0: in Radius; R: out Radius; Found: out Boolean) is
--- Find a root >= R0 of C0+R*(C1+R*R*(C3+R))
    Eps: constant Radius := 1.0E-10;
    Rmin,Rmax,Y: Rep;
  begin
    Found := False;
    R := R0;
    if (C0+R*(C1+R*R*(C3+R))) >= Zero then return; end if;
    Rmin := R0;
    Rmax := R0+One;
    while (C0+Rmax*(C1+Rmax*Rmax*(C3+Rmax)))<Zero loop Rmax := Two*Rmax; end loop;
    loop
      R := Half*(Rmin+Rmax);
      Y := C0+R*(C1+R*R*(C3+R));
      exit when abs(Y)<Eps;
      if Y>Zero then Rmax := R; else Rmin := R; end if;
    end loop;
    Found := True;
  end Root4;

end Reps.Ops;
