with Ada.Text_IO, Messages, Reps.Ops, Linear.Ops;
use Ada.Text_IO, Messages, Reps.Ops, Linear.Ops;

pragma Elaborate_All (Ada.Text_IO, Reps.Ops, Linear.Ops);

package body SLinear.Ops is

  ScalZero: constant Scalar := Scal(Zero);
  ScalOne: constant Scalar := Scal(One);
  ScalTwo: constant Scalar := Scal(Two);

  function Approx(A: SMatrix) return Matrix is
    B: Matrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        B(I,J) := Approx(A(I,J));
      end loop;
    end loop;
    return B;
  end Approx;

  function GetNumeric(A: Matrix) return SMatrix is
    B: SMatrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        B(I,J) := Scal(A(I,J));
      end loop;
    end loop;
    return B;
  end GetNumeric;

  procedure Add(B1: in SVector; B2: in out SVector) is
  begin
    for I in B2'Range loop
      B2(I) := B2(I)+B1(I);
    end loop;
  end Add;

  procedure Add(A1: in SMatrix; A2: in out SMatrix) is
  begin
    for I in A2'Range(1) loop
      for J in A2'Range(2) loop
         A2(I,J) := A2(I,J)+A1(I,J);
      end loop;
    end loop;
  end Add;

  procedure Add(R: in Scalar; A: in out SMatrix) is
  begin
    for I in A'Range(1) loop
      A(I,I) := A(I,I)+R;
    end loop;
  end Add;

  procedure Sub(B1: in SVector; B2: in out SVector) is
  begin
    for I in B2'Range loop
      B2(I) := B2(I)-B1(I);
    end loop;
  end Sub;

  procedure Sub(A1: in SMatrix; A2: in out SMatrix) is
  begin
    for I in A2'Range(1) loop
      for J in A2'Range(2) loop
         A2(I,J) := A2(I,J)-A1(I,J);
      end loop;
    end loop;
  end Sub;

  procedure Mult(R: in Scalar; B: in out SVector) is
  begin
    for I in B'Range loop
      B(I) := R*B(I);
    end loop;
  end Mult;

  procedure Mult(R: in Scalar; A: in out SMatrix) is
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        A(I,J) := R*A(I,J);
      end loop;
    end loop;
  end Mult;

   procedure Mult(A1: in SMatrix; A2: in out SMatrix) is
    R: Scalar;
    B: SVector(A2'Range(1));
  begin
    for J in B'Range loop
      for K in B'Range loop
        B(K) := A2(K,J);
      end loop;
      for I in B'Range loop
        R := ScalZero;
        for K in B'Range loop
          R := R+A1(I,K)*B(K);
        end loop;
        A2(I,J) := R;
      end loop;
    end loop;
  end Mult;

  function "-"(B: SVector) return SVector is
    Bt: SVector(B'Range);
  begin
    for I in B'Range loop
      Bt(I) := -B(I);
    end loop;
    return Bt;
   end "-";

  function "+"(B1,B2: SVector) return SVector is
    B: SVector(B2'Range);
  begin
    for I in B'Range loop
      B(I) := B2(I)+B1(I);
    end loop;
    return B;
  end "+";

  function "+"(A1,A2: in SMatrix) return SMatrix is
    A: SMatrix(A2'Range(1),A2'Range(2));
  begin
    for I in A'Range loop
      for J in A'Range loop
         A(I,J) := A2(I,J)+A1(I,J);
      end loop;
    end loop;
    return A;
  end "+";

  function "-"(B1,B2: SVector) return SVector is
    B: SVector(B2'Range);
  begin
    for I in B'Range loop
      B(I) := B2(I)-B1(I);
    end loop;
    return B;
  end "-";

  function "-"(A1,A2: in SMatrix) return SMatrix is
    A: SMatrix(A2'Range(1),A2'Range(2));
  begin
    for I in A'Range loop
      for J in A'Range loop
         A(I,J) := A2(I,J)-A1(I,J);
      end loop;
    end loop;
    return A;
  end "-";

  function "*"(B1,B2: SVector) return Scalar is
    R: Scalar;
  begin
    R := ScalZero;
    for I in B2'Range loop
      R := R+B1(I)*B2(I);
    end loop;
    return R;
  end "*";

  function "*"(R: Scalar; B: SVector) return SVector is
    Br: SVector(B'Range);
  begin
    for I in B'Range loop
      Br(I) := R*B(I);
    end loop;
    return Br;
  end "*";

  function "*"(R: Scalar; A: SMatrix) return SMatrix is
    Ar: SMatrix(A'Range(1),A'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        Ar(I,J) := R*A(I,J);
      end loop;
    end loop;
    return Ar;
  end "*";

  function "*"(B1,B2: SVector) return SVector is
    B: SVector(B2'Range);
  begin
    for I in B2'Range loop
      B(I) := B1(I)*B2(I);
    end loop;
    return B;
  end "*";

  function "*"(A: SMatrix; B: SVector) return SVector is
    R: Scalar;
    Br: SVector(A'Range(1));
  begin
    for I in Br'Range loop
      R := ScalZero;
      for J in B'Range loop
        R := R+A(I,J)*B(J);
      end loop;
      Br(I) := R;
    end loop;
    return Br;
  end "*";

  function "*"(B: SVector; A: SMatrix) return SVector is
    R: Scalar;
    Br: SVector(A'Range(2));
  begin
    for I in Br'Range loop
      R := ScalZero;
      for J in B'Range loop
        R := R+B(J)*A(J,I);
      end loop;
      Br(I) := R;
    end loop;
    return Br;
  end "*";

  function "*"(A1,A2: SMatrix) return SMatrix is
    R: Scalar;
    A: SMatrix(A1'Range(1),A2'Range(2));
  begin
    for I in A'Range(1) loop
      for J in A'Range(2) loop
        R := ScalZero;
        for K in A2'Range(1) loop
          R := R+A1(I,K)*A2(K,J);
        end loop;
        A(I,J) := R;
      end loop;
    end loop;
    return A;
  end "*";

  procedure SwapRow(I1,I2: in Integer; A: in out SMatrix) is
    S: Scalar;
  begin
    if I1=I2 then return; end if;
    for J in A'Range(2) loop
      S := A(I1,J);
      A(I1,J) := A(I2,J);
      A(I2,J) := S;
    end loop;
  end;

  procedure Invert(A: in out SMatrix) is
    Matrix_Singular: exception;
    N: constant Integer := A'Last(1);
    Eps: constant Rep := (One+Rep'Safe_Small)-One;
    L,M,Pt: Integer;
    R: Rep;
    S: Scalar;
    P: Pivot(1 .. N);
    C: Vector(1 .. N);
    B: SMatrix(1 .. N,1 .. N);

    function Quality(X: Scalar) return Rep is
      Xi: constant Rep := InfAbs(X);
      Xs: constant Rep := SupAbs(X);
    begin
      if Xi=Xs then return Xi/Sqrt(Eps); end if;
      return Xi/Sqrt(One+Eps-(Xi/Xs));
    end Quality;

    procedure FindBest(K: in Integer; I0,J0: out Integer) is
      Q: Rep := Zero;
    begin
      for I in K .. N loop
        for J in K .. N loop
          R := Quality(A(I,J));
          if R>Q then Q := R; I0 := I; J0 := J; end if;
        end loop;
      end loop;
      if Q=Zero then raise Matrix_Singular; end if;
    end FindBest;

    procedure SwapCol(J1,J2: in Integer) is
    begin
      if J1=J2 then return; end if;
      for I in A'Range(1) loop
        S := A(I,J1);
        A(I,J1) := A(I,J2);
        A(I,J2) := S;
      end loop;
      Pt := P(J1); P(J1) := P(J2); P(J2) := Pt;
    end SwapCol;

  begin
    for J in 1 .. N loop               -- B := A*diag(C), A := Id
      R := Zero;
      for I in 1 .. N loop
        R := Rep'Max(R,abs(Approx(A(I,J))));
      end loop;
      R := Rep'Leading_Part(One/R,1);  -- a power of Two
      for I in 1 .. N loop
        A(I,J) := R*A(I,J);
        B(I,J) := ScalZero;
      end loop;
      B(J,J) := ScalOne;
      C(J) := R;
      P(J) := J;
    end loop;

    for K in 1 .. N loop               -- Gaussian
      FindBest(K,L,M);
      SwapRow(K,L,A);
      SwapRow(K,L,B);
      SwapCol(K,M);
      S := Inv(A(K,K));
      A(K,K) := ScalOne;
      for J in K+1 .. N loop
        A(K,J) := S*A(K,J);
      end loop;
      for J in 1 .. N loop
        B(K,J) := S*B(K,J);
      end loop;

      for I in 1 .. N loop
        if I /= K then
          S := A(I,K);
          A(I,K) := ScalZero;
          for J in K+1 .. N loop
            A(I,J) := A(I,J)-S*A(K,J);
          end loop;
          for J in 1 .. N loop
            B(I,J) := B(I,J)-S*B(K,J);
          end loop;
        end if;
      end loop;
    end loop;

    for I in 1 .. N loop
      Pt := P(I);
      R := C(Pt);
      for J in 1 .. N loop
        A(Pt,J) := R*B(I,J);
      end loop;
    end loop;

  end Invert;

  procedure Dgefa(A: in out SMatrix; Pvt: out Pivot) is
--- adapted from linpack routine dgefa, http://www.netlib.org/linpack/
--- factors a matrix by gaussian elimination, see Linear.Ops.dgefa for details
    N: constant Integer := Pvt'Last;
    L: Integer;
    X,Y: Radius;
    T: Scalar;
  begin
    for K in 1 .. N-1 loop
      Y := Inf(abs(A(N,K)));
      L := N;
      for M in reverse K .. N-1 loop
        X := Inf(abs(A(M,K)));
        if X >= Y then Y := X; L := M; end if;
      end loop;
      Pvt(K) := L;
      if A(L,K) /= ScalZero then
        if L /= K then
          T := A(L,K);
          A(L,K) := A(K,K);
          A(K,K) := T;
        end if;
        T := -Inv(A(K,K));
        for M in K+1 .. N loop
          A(M,K) := T*A(M,K);
        end loop;
        for J in K+1 .. N loop
          T := A(L,J);
          if L /= K then
            A(L,J) := A(K,J);
            A(K,J) := T;
          end if;
          for M in K+1 .. N loop
            A(M,J) := A(M,J)+T*A(M,K);
          end loop;
        end loop;
      end if;
    end loop;
    Pvt(N) := N;
  end Dgefa;

  procedure Dgesl(Pvt: in Pivot; A: in SMatrix; B: in out SVector) is
--- adapted from linpack routine dgesl, http://www.netlib.org/linpack/
--- solves the system AX=B, see Linear.Ops.dgesl for details
    N: constant Integer := Pvt'Last;
    L: Integer;
    T: Scalar;
  begin
    for K in 1 .. N-1 loop
      L := Pvt(K);
      T := B(L);
      if L /= K then
        B(L) := B(K);
        B(K) := T;
      end if;
      for M in K+1 .. N loop
        B(M) := B(M)+T*A(M,K);
      end loop;
    end loop;
    for K in reverse 1 .. N loop
      B(K) := B(K)/A(K,K);
      T := -B(K);
      for M in 1 .. K-1 loop
        B(M) := B(M)+T*A(M,K);
      end loop;
    end loop;
  end Dgesl;

  procedure Dgedi(Pvt: in Pivot; A: in out SMatrix; B: in out SVector) is
--- adapted from linpack routine dgedi, http://www.netlib.org/linpack/
--- computes the inverse of a matrix, see Linear.Ops.dgedi for details
    N: constant Integer := Pvt'Last;
    L: integer;
    T: Scalar;
  begin
    for K in 1 .. N loop
      A(K,K) := Inv(A(K,K));
      T := -A(K,K);
      for M in 1 .. K-1 loop
        A(M,K) := T*A(M,K);
      end loop;
      for J in K+1 .. N loop
        T := A(K,J);
        A(K,J) := ScalZero;
        for M in 1 .. K loop
          A(M,J) := A(M,J)+T*A(M,K);
        end loop;
      end loop;
    end loop;
    for K in reverse 1 .. N-1 loop
      for I in K+1 .. N loop
        B(I) := A(I,K);
        A(I,K) := ScalZero;
      end loop;
      for J in K+1 .. N loop
        T := B(J);
        for M in 1 .. N loop
          A(M,K) := A(M,K)+T*A(M,J);
        end loop;
      end loop;
      L := Pvt(K);
      if L /= K  then
        for M in 1 .. N loop
          T := A(M,K);
          A(M,K) := A(M,L);
          A(M,L) := T;
        end loop;
      end if;
    end loop;
  end Dgedi;

  procedure Solve(A: in SMatrix; B: in SVector; X: out SVector) is
    Pvt: Pivot(A'Range(1));
    Ar: SMatrix(A'Range(1),A'Range(2));
  begin
    Ar := A;
    X := B;
    Dgefa(Ar,Pvt);
    Dgesl(Pvt,Ar,X);
  end Solve;

  procedure LPInvert(A: in out SMatrix) is
    Pvt: Pivot(A'Range(2));
    B: SVector(A'Range(2));
  begin
    Dgefa(A,Pvt);
    Dgedi(Pvt,A,B);
  end LPInvert;

end SLinear.Ops;
