with Linear;
use Linear;

generic

  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with function "*"(R: Rep; S: Scalar) return Scalar is <>;
  with function "/"(S: Scalar; R: Rep) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "/"(S,T: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;

package SLinear.Ops is

  function Approx(A: SMatrix) return Matrix;                          -- approximate SMatrix
  function GetNumeric(A: Matrix) return SMatrix;                      -- convert to SMatrix

  procedure Add(B1: in SVector; B2: in out SVector);                  -- B2 := B2+B1
  procedure Add(A1: in SMatrix; A2: in out SMatrix);                  -- A2 := A2+A1
  procedure Add(R: in Scalar; A: in out SMatrix);                     -- A := A+R*Identity
  procedure Sub(B1: in SVector; B2: in out SVector);                  -- B2 := B2-B1
  procedure Sub(A1: in SMatrix; A2: in out SMatrix);                  -- A2 := A2-A1
  procedure Mult(R: in Scalar; B: in out SVector);                    -- B := R*B
  procedure Mult(R: in Scalar; A: in out SMatrix);                    -- A := R*A
  procedure Mult(A1: in SMatrix; A2: in out SMatrix);                 -- A2 := A1*A2

  function "-"(B: SVector) return SVector;                            -- -B
  function "+"(B1,B2: SVector) return SVector;                        -- B1+B2
  function "+"(A1,A2: in SMatrix) return SMatrix;                     -- A1+A2
  function "-"(B1,B2: SVector) return SVector;                        -- B1-B2
  function "-"(A1,A2: in SMatrix) return SMatrix;                     -- A1-A2
  function "*"(B1,B2: SVector) return Scalar;                         -- L2 inner product <B1,B2>
  function "*"(R: Scalar; B: SVector) return SVector;                 -- R*B
  function "*"(R: Scalar; A: SMatrix) return SMatrix;                 -- R*A
  function "*"(B1,B2: SVector) return SVector;                        -- Diag(B1)*B2
  function "*"(A: SMatrix; B: SVector) return SVector;                -- A*B
  function "*"(B: SVector; A: SMatrix) return SVector;                -- Transpose(A)*B
  function "*"(A1,A2: SMatrix) return SMatrix;                        -- A1*A2

  procedure Invert(A: in out SMatrix);                                -- Invert A
  procedure Solve(A: in SMatrix; B: in SVector; X: out SVector);      -- Solve A*X=B
  procedure LPInvert(A: in out SMatrix);                              -- Invert A

private

  procedure SwapRow(I1,I2: in Integer; A: in out SMatrix);
  procedure Dgefa(A: in out SMatrix; Pvt: out Pivot);
  procedure Dgesl(Pvt: in Pivot; A: in SMatrix; B: in out SVector);
  procedure Dgedi(Pvt: in Pivot; A: in out SMatrix; B: in out SVector);

end SLinear.Ops;
